#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict, List, NamedTuple, Optional, Union

from . import ClassyParamScheduler, register_param_scheduler


@register_param_scheduler("step")
class StepParamScheduler(ClassyParamScheduler):
    """
    Takes a fixed schedule for a param value.  If the length of the
    fixed schedule is less than the number of epochs, then the epochs
    are divided evenly among the param schedule.

    Example:

        .. code-block:: python

          values: [0.1, 0.01, 0.001, 0.0001]
          num_epochs = 120

    Then the param value will be 0.1 for epochs 0-29, 0.01 for
    epochs 30-59, 0.001 for epoch 60-89, 0.0001 for epochs 90-119.
    """

    def __init__(self, num_epochs: Union[int, float], values: List[float]):
        super().__init__()

        self._param_schedule = values

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "StepParamScheduler":
        """Instantiates a StepParamScheduler from a configuration.

        Args:
            config: A configuration for a StepParamScheduler.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A StepParamScheduler instance.
        """
        assert (
            "values" in config
            and isinstance(config["values"], list)
            and len(config["values"]) > 0
        ), "Step scheduler requires a list of at least one param value"
        assert config["num_epochs"] > 0, "Num epochs must be greater than 0"

        return cls(num_epochs=config["num_epochs"], values=config["values"])

    def __call__(self, where: float):
        ind = int((where + self.WHERE_EPSILON) * len(self._param_schedule))
        return self._param_schedule[ind]



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict

from . import ClassyParamScheduler, register_param_scheduler


@register_param_scheduler("linear")
class LinearParamScheduler(ClassyParamScheduler):
    """
    Linearly interpolates parameter between ``start_lr`` and ``end_lr``.
    Can be used for either warmup or decay based on start and end values.

    Example:

        .. code-block:: python

            start_lr: 0.0001
            end_lr: 0.01
    Corresponds to a linear increasing schedule with values in [0.0001, 0.01)
    """

    def __init__(self, start_lr: float, end_lr: float):
        super().__init__()
        self._start_lr = start_lr
        self._end_lr = end_lr

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "LinearParamScheduler":
        """Instantiates a LinearParamScheduler from a configuration.

        Args:
            config: A configuration for a LinearParamScheduler.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A LinearParamScheduler instance.
        """
        assert (
            "start_lr" in config and "end_lr" in config
        ), "Linear scheduler requires a start and a end"
        return cls(start_lr=config["start_lr"], end_lr=config["end_lr"])

    def __call__(self, where: float):
        # interpolate between start and end values
        return self._end_lr * where + self._start_lr * (1 - where)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict

import torch.optim

from . import ClassyOptimizer, register_optimizer


@register_optimizer("sgd")
class SGD(ClassyOptimizer):
    def __init__(
        self,
        lr: float = 0.1,
        momentum: float = 0.0,
        weight_decay: float = 0.0,
        nesterov: bool = False,
    ):
        super().__init__()

        self.parameters.lr = lr
        self.parameters.momentum = momentum
        self.parameters.weight_decay = weight_decay
        self.parameters.nesterov = nesterov

    def init_pytorch_optimizer(self, model, **kwargs):
        super().init_pytorch_optimizer(model, **kwargs)
        self.optimizer = torch.optim.SGD(
            self.param_groups_override,
            lr=self.parameters.lr,
            nesterov=self.parameters.nesterov,
            momentum=self.parameters.momentum,
            weight_decay=self.parameters.weight_decay,
        )

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "SGD":
        """Instantiates a SGD from a configuration.

        Args:
            config: A configuration for a SGD.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A SGD instance.
        """
        # Default params
        config.setdefault("lr", 0.1)
        config.setdefault("momentum", 0.0)
        config.setdefault("weight_decay", 0.0)
        config.setdefault("nesterov", False)

        assert (
            config["momentum"] >= 0.0
            and config["momentum"] < 1.0
            and type(config["momentum"]) == float
        ), "Config must contain a 'momentum' in [0, 1) for SGD optimizer"
        assert isinstance(
            config["nesterov"], bool
        ), "Config must contain a boolean 'nesterov' param for SGD optimizer"

        return cls(
            lr=config["lr"],
            momentum=config["momentum"],
            weight_decay=config["weight_decay"],
            nesterov=config["nesterov"],
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from abc import ABC, abstractmethod
from unittest.mock import MagicMock

import torch
from classy_vision.models import ClassyModel
from classy_vision.optim import build_optimizer


class TestOptimizer(ABC):
    @abstractmethod
    def _get_config(self):
        pass

    @abstractmethod
    def _instance_to_test(self):
        pass

    def _check_momentum_buffer(self):
        return False

    def _get_optimizer_params(self):
        return {
            "regularized_params": [
                torch.tensor([[1.0, 2.0]], requires_grad=True),
                torch.tensor([[3.0, 4.0]], requires_grad=True),
            ],
            "unregularized_params": [torch.tensor([[1.0, 2.0]], requires_grad=True)],
        }

    def _get_mock_classy_vision_model(self, trainable_params=True):
        mock_classy_vision_model = ClassyModel()

        if trainable_params:
            mock_classy_vision_model.get_optimizer_params = MagicMock(
                return_value=self._get_optimizer_params()
            )
            mock_classy_vision_model.parameters = MagicMock(
                return_value=self._get_optimizer_params()["regularized_params"]
                + self._get_optimizer_params()["unregularized_params"]
            )
        else:
            mock_classy_vision_model.get_optimizer_params = MagicMock(
                return_value={"regularized_params": [], "unregularized_params": []}
            )
            mock_classy_vision_model.parameters = MagicMock(
                return_value=[
                    param.detach()
                    for param in self._get_optimizer_params()["regularized_params"]
                    + self._get_optimizer_params()["unregularized_params"]
                ]
            )

        return mock_classy_vision_model

    def _set_gradient(self, params, grad_values=None):
        if grad_values is None:
            grad_values = [0.1, 0.1]
        for i in range(len(params)):
            params[i].grad = torch.tensor([grad_values])

    def _set_model_gradient(self, model, grad_values=None):
        for param_type in ["regularized_params", "unregularized_params"]:
            self._set_gradient(model.get_optimizer_params()[param_type], grad_values)

    def _compare_momentum_values(self, optim1, optim2):
        self.assertEqual(len(optim1["param_groups"]), len(optim2["param_groups"]))

        for i in range(len(optim1["param_groups"])):
            self.assertEqual(
                len(optim1["param_groups"][i]["params"]),
                len(optim2["param_groups"][i]["params"]),
            )
            if self._check_momentum_buffer():
                for j in range(len(optim1["param_groups"][i]["params"])):
                    id1 = optim1["param_groups"][i]["params"][j]
                    id2 = optim2["param_groups"][i]["params"][j]
                    self.assertTrue(
                        torch.allclose(
                            optim1["state"][id1]["momentum_buffer"],
                            optim2["state"][id2]["momentum_buffer"],
                        )
                    )

    def _get_set_state(self, grad_values):
        config = self._get_config()

        mock_classy_vision_model = self._get_mock_classy_vision_model()
        opt1 = build_optimizer(config)
        opt1.init_pytorch_optimizer(mock_classy_vision_model)

        self._set_model_gradient(mock_classy_vision_model, grad_values)
        opt1.step()
        state = opt1.get_classy_state()

        config["lr"] += 0.1
        opt2 = build_optimizer(config)
        opt2.init_pytorch_optimizer(mock_classy_vision_model)
        self.assertTrue(isinstance(opt1, self._instance_to_test()))
        opt2.set_classy_state(state)
        self.assertEqual(opt1.parameters, opt2.parameters)
        for i in range(len(opt1.optimizer.param_groups[0]["params"])):
            self.assertTrue(
                torch.allclose(
                    opt1.optimizer.param_groups[0]["params"][i],
                    opt2.optimizer.param_groups[0]["params"][i],
                )
            )
        self._compare_momentum_values(
            opt1.get_classy_state()["optim"], opt2.get_classy_state()["optim"]
        )

        # check if the optimizers behave the same on params update
        mock_classy_vision_model1 = self._get_mock_classy_vision_model()
        mock_classy_vision_model2 = self._get_mock_classy_vision_model()
        self._set_model_gradient(mock_classy_vision_model1, grad_values)
        self._set_model_gradient(mock_classy_vision_model2, grad_values)
        opt1 = build_optimizer(config)
        opt1.init_pytorch_optimizer(mock_classy_vision_model1)
        opt2 = build_optimizer(config)
        opt2.init_pytorch_optimizer(mock_classy_vision_model2)
        opt1.step()
        opt2.step()
        for i in range(len(opt1.optimizer.param_groups[0]["params"])):
            print(opt1.optimizer.param_groups[0]["params"][i])
            self.assertTrue(
                torch.allclose(
                    opt1.optimizer.param_groups[0]["params"][i],
                    opt2.optimizer.param_groups[0]["params"][i],
                )
            )
        self._compare_momentum_values(
            opt1.get_classy_state()["optim"], opt2.get_classy_state()["optim"]
        )

    def test_build_sgd(self):
        config = self._get_config()
        mock_classy_vision_model = self._get_mock_classy_vision_model(
            trainable_params=True
        )
        opt = build_optimizer(config)
        opt.init_pytorch_optimizer(mock_classy_vision_model)
        self.assertTrue(isinstance(opt, self._instance_to_test()))

    def test_raise_error_on_non_trainable_params(self):
        # Test Raise ValueError if there are no trainable params in the model.
        config = self._get_config()
        with self.assertRaises(ValueError):
            opt = build_optimizer(config)
            opt.init_pytorch_optimizer(
                self._get_mock_classy_vision_model(trainable_params=False)
            )

    def test_get_set_state(self):
        for grad_values in [[0.1, 0.1], [-0.1, -0.1], [0.0, 0.0], [0.1, -0.1]]:
            self._get_set_state(grad_values)

    def test_set_invalid_state(self):
        config = self._get_config()
        mock_classy_vision_model = self._get_mock_classy_vision_model()
        opt = build_optimizer(config)
        opt.init_pytorch_optimizer(mock_classy_vision_model)
        self.assertTrue(isinstance(opt, self._instance_to_test()))

        with self.assertRaises(KeyError):
            opt.set_classy_state({})

    def test_lr_schedule(self):
        config = self._get_config()

        mock_classy_vision_model = self._get_mock_classy_vision_model()
        opt = build_optimizer(config)
        opt.init_pytorch_optimizer(mock_classy_vision_model)

        # Test initial learning rate
        for group in opt.optimizer.param_groups:
            self.assertEqual(group["lr"], 0.1)

        def _test_lr_schedule(optimizer, num_epochs, epochs, targets):
            for i in range(len(epochs)):
                epoch = epochs[i]
                target = targets[i]
                param_groups = optimizer.optimizer.param_groups.copy()
                optimizer.update_schedule_on_epoch(epoch / num_epochs)
                for idx, group in enumerate(optimizer.optimizer.param_groups):
                    self.assertEqual(group["lr"], target)
                    # Make sure all but LR is same
                    param_groups[idx]["lr"] = target
                    self.assertEqual(param_groups[idx], group)

        # Test constant learning schedule
        num_epochs = 90
        epochs = [0, 0.025, 0.05, 0.1, 0.5, 1, 15, 29, 30, 31, 59, 60, 61, 88, 89]
        targets = [0.1] * 15
        _test_lr_schedule(opt, num_epochs, epochs, targets)

        # Test step learning schedule
        config["lr"] = {"name": "step", "values": [0.1, 0.01, 0.001]}
        opt = build_optimizer(config)
        opt.init_pytorch_optimizer(mock_classy_vision_model)
        targets = [0.1] * 8 + [0.01] * 3 + [0.001] * 4
        _test_lr_schedule(opt, num_epochs, epochs, targets)

        # Test step learning schedule with warmup
        init_lr = 0.01
        warmup_epochs = 0.1
        config["lr"] = {
            "name": "composite",
            "schedulers": [
                {"name": "linear", "start_lr": init_lr, "end_lr": 0.1},
                {"name": "step", "values": [0.1, 0.01, 0.001]},
            ],
            "update_interval": "epoch",
            "interval_scaling": ["rescaled", "fixed"],
            "lengths": [warmup_epochs / num_epochs, 1 - warmup_epochs / num_epochs],
        }

        opt = build_optimizer(config)
        opt.init_pytorch_optimizer(mock_classy_vision_model)
        targets = [0.01, 0.0325, 0.055] + [0.1] * 5 + [0.01] * 3 + [0.001] * 4
        _test_lr_schedule(opt, num_epochs, epochs, targets)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict

import torch

from . import ClassyLoss, register_loss


@register_loss("barron")
class BarronLoss(ClassyLoss):
    """
    This implements the `Barron loss <https://arxiv.org/pdf/1701.03077.pdf>`_.
    """

    def __init__(self, alpha, size_average, c):
        super(BarronLoss, self).__init__()

        self.size_average = size_average
        self.alpha = alpha
        self.c = c
        self.z = max(1.0, 2.0 - self.alpha)

        # define all three losses:
        def _forward_zero(diff):
            out = diff.div(self.c).pow(2.0).mul(0.5).add(1.0).log()
            return out

        def _forward_inf(diff):
            out = 1.0 - diff.div(self.c).pow(2.0).mul(-0.5).exp()
            return out

        def _forward(diff):
            out = diff.div(self.c).pow(2.0).div(self.z).add(1.0).pow(self.alpha / 2.0)
            out.add_(-1.0).mul_(self.z / self.alpha)
            return out

        # set the correct loss:
        if self.alpha == 0.0:
            self._forward = _forward_zero
        elif self.alpha == -float("inf") or self.alpha == float("inf"):
            self._forward = _forward_inf
        else:
            self._forward = _forward

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "BarronLoss":
        """Instantiates a BarronLoss from a configuration.

        Args:
            config: A configuration for a BarronLoss.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A BarronLoss instance.
        """
        # Infinity is a valid alpha value but is frequently a string
        config["alpha"] = float(config["alpha"])
        # assertions:
        assert type(config["size_average"]) == bool
        assert type(config["alpha"]) == float
        assert type(config["c"]) == float and config["c"] > 0.0

        return cls(
            alpha=config["alpha"], size_average=config["size_average"], c=config["c"]
        )

    def forward(self, prediction, target):
        diff = torch.add(prediction, -target)
        loss = self._forward(diff)
        loss = loss.sum(0, keepdim=True)
        if self.size_average:
            loss.div_(prediction.size(0))
        return loss



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch


# Default to GPU 0
_cuda_device_index: int = 0

# Setting _cuda_device_index to -1 internally implies that we should use CPU
_CPU_DEVICE_INDEX = -1


def convert_to_distributed_tensor(tensor):
    """
    For some backends, such as NCCL, communication only works if the
    tensor is on the GPU. This helper function converts to the correct
    device and returns the tensor + original device.
    """
    orig_device = "cpu" if not tensor.is_cuda else "gpu"
    if (
        torch.distributed.is_available()
        and torch.distributed.get_backend() == torch.distributed.Backend.NCCL
        and not tensor.is_cuda
    ):
        tensor = tensor.cuda()
    return (tensor, orig_device)


def convert_to_normal_tensor(tensor, orig_device):
    """
    For some backends, such as NCCL, communication only works if the
    tensor is on the GPU. This converts the tensor back to original device.
    """
    if tensor.is_cuda and orig_device == "cpu":
        tensor = tensor.cpu()
    return tensor


def is_distributed_training_run():
    return (
        torch.distributed.is_available()
        and torch.distributed.is_initialized()
        and (torch.distributed.get_world_size() > 1)
    )


def is_master():
    """
    Returns True if this is rank 0 of a distributed training job OR if it is
    a single trainer job. Otherwise False.
    """
    return get_rank() == 0


def all_reduce_mean(tensor):
    """
    Wrapper over torch.distributed.all_reduce for performing mean reduction
    of tensor over all processes.
    """
    if is_distributed_training_run():
        tensor, orig_device = convert_to_distributed_tensor(tensor)
        torch.distributed.all_reduce(tensor, torch.distributed.ReduceOp.SUM)
        tensor = tensor / torch.distributed.get_world_size()
        tensor = convert_to_normal_tensor(tensor, orig_device)
    return tensor


def all_reduce_sum(tensor):
    """
    Wrapper over torch.distributed.all_reduce for performing sum
    reduction of tensor over all processes in both distributed /
    non-distributed scenarios.
    """
    if is_distributed_training_run():
        tensor, orig_device = convert_to_distributed_tensor(tensor)
        torch.distributed.all_reduce(tensor, torch.distributed.ReduceOp.SUM)
        tensor = convert_to_normal_tensor(tensor, orig_device)
    return tensor


def gather_tensors_from_all(tensor):
    """
    Wrapper over torch.distributed.all_gather for performing
    'gather' of 'tensor' over all processes in both distributed /
    non-distributed scenarios.
    """
    if tensor.ndim == 0:
        # 0 dim tensors cannot be gathered. so unsqueeze
        tensor = tensor.unsqueeze(0)

    if is_distributed_training_run():
        tensor, orig_device = convert_to_distributed_tensor(tensor)
        gathered_tensors = [
            torch.zeros_like(tensor) for _ in range(torch.distributed.get_world_size())
        ]
        torch.distributed.all_gather(gathered_tensors, tensor)
        gathered_tensors = [
            convert_to_normal_tensor(_tensor, orig_device)
            for _tensor in gathered_tensors
        ]
    else:
        gathered_tensors = [tensor]

    return gathered_tensors


def gather_from_all(tensor):
    gathered_tensors = gather_tensors_from_all(tensor)
    gathered_tensor = torch.cat(gathered_tensors, 0)
    return gathered_tensor


def barrier():
    """
    Wrapper over torch.distributed.barrier, returns without waiting
    if the distributed process group is not initialized instead of throwing error.
    """
    if not torch.distributed.is_available() or not torch.distributed.is_initialized():
        return
    torch.distributed.barrier()


def get_world_size():
    """
    Simple wrapper for correctly getting worldsize in both distributed
    / non-distributed settings
    """
    return (
        torch.distributed.get_world_size()
        if torch.distributed.is_available() and torch.distributed.is_initialized()
        else 1
    )


def get_rank():
    """
    Simple wrapper for correctly getting rank in both distributed
    / non-distributed settings
    """
    return (
        torch.distributed.get_rank()
        if torch.distributed.is_available() and torch.distributed.is_initialized()
        else 0
    )


def set_cuda_device_index(idx: int):
    global _cuda_device_index
    _cuda_device_index = idx
    torch.cuda.set_device(_cuda_device_index)


def set_cpu_device():
    global _cuda_device_index
    _cuda_device_index = _CPU_DEVICE_INDEX


def get_cuda_device_index() -> int:
    return _cuda_device_index


def init_distributed_data_parallel_model(model, broadcast_buffers=False):
    global _cuda_device_index

    if _cuda_device_index == _CPU_DEVICE_INDEX:
        # CPU-only model, don't specify device
        return torch.nn.parallel.DistributedDataParallel(
            model, broadcast_buffers=broadcast_buffers
        )
    else:
        # GPU model
        return torch.nn.parallel.DistributedDataParallel(
            model,
            device_ids=[_cuda_device_index],
            output_device=_cuda_device_index,
            broadcast_buffers=broadcast_buffers,
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from pathlib import Path

from classy_vision.generic.registry_utils import import_all_modules


# The order of imports matters here because of a circular dependency. constants
# must come before any hook
from .constants import ClassyHookFunctions  # isort:skip
from .checkpoint_hook import CheckpointHook  # isort:skip
from .classy_hook import ClassyHook  # isort:skip
from .exponential_moving_average_model_hook import (  # isort:skip
    ExponentialMovingAverageModelHook,
)
from .loss_lr_meter_logging_hook import LossLrMeterLoggingHook  # isort:skip
from .model_complexity_hook import ModelComplexityHook  # isort:skip
from .model_tensorboard_hook import ModelTensorboardHook  # isort:skip
from .profiler_hook import ProfilerHook  # isort:skip
from .progress_bar_hook import ProgressBarHook  # isort:skip
from .tensorboard_plot_hook import TensorboardPlotHook  # isort:skip
from .time_metrics_hook import TimeMetricsHook  # isort:skip
from .visdom_hook import VisdomHook  # isort:skip


__all__ = [
    "CheckpointHook",
    "ClassyHook",
    "ClassyHookFunctions",
    "ExponentialMovingAverageModelHook",
    "LossLrMeterLoggingHook",
    "TensorboardPlotHook",
    "ModelComplexityHook",
    "ModelTensorboardHook",
    "ProfilerHook",
    "ProgressBarHook",
    "TimeMetricsHook",
    "VisdomHook",
]

FILE_ROOT = Path(__file__).parent

# automatically import any Python files in the hooks/ directory
import_all_modules(FILE_ROOT, "classy_vision.hooks")



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import unittest

from classy_vision.optim.param_scheduler import build_param_scheduler
from classy_vision.optim.param_scheduler.multi_step_scheduler import (
    MultiStepParamScheduler,
)


class TestMultiStepParamScheduler(unittest.TestCase):
    _num_epochs = 12

    def _get_valid_config(self):
        return {
            "name": "multistep",
            "num_epochs": self._num_epochs,
            "values": [0.1, 0.01, 0.001, 0.0001],
            "milestones": [4, 6, 8],
        }

    def test_invalid_config(self):
        # Invalid num epochs
        config = self._get_valid_config()

        bad_config = copy.deepcopy(config)
        bad_config["num_epochs"] = -1
        with self.assertRaises(AssertionError):
            MultiStepParamScheduler.from_config(bad_config)

        # Invalid values
        bad_config["num_epochs"] = config["num_epochs"]
        del bad_config["values"]
        with self.assertRaises(AssertionError):
            MultiStepParamScheduler.from_config(bad_config)

        bad_config["values"] = {"a": "b"}
        with self.assertRaises(AssertionError):
            MultiStepParamScheduler.from_config(bad_config)

        bad_config["values"] = []
        with self.assertRaises(AssertionError):
            MultiStepParamScheduler.from_config(bad_config)

        # Invalid drop epochs
        bad_config["values"] = config["values"]
        bad_config["milestones"] = {"a": "b"}
        with self.assertRaises(AssertionError):
            MultiStepParamScheduler.from_config(bad_config)

        # Too many
        bad_config["milestones"] = [3, 6, 8, 12]
        with self.assertRaises(AssertionError):
            MultiStepParamScheduler.from_config(bad_config)

        # Too few
        bad_config["milestones"] = [3, 6]
        with self.assertRaises(AssertionError):
            MultiStepParamScheduler.from_config(bad_config)

        # Exceeds num_epochs
        bad_config["milestones"] = [3, 6, 12]
        with self.assertRaises(AssertionError):
            MultiStepParamScheduler.from_config(bad_config)

        # Out of order
        bad_config["milestones"] = [3, 8, 6]
        with self.assertRaises(AssertionError):
            MultiStepParamScheduler.from_config(bad_config)

    def _test_config_scheduler(self, config, expected_schedule):
        scheduler = MultiStepParamScheduler.from_config(config)
        schedule = [
            scheduler(epoch_num / self._num_epochs)
            for epoch_num in range(self._num_epochs)
        ]
        self.assertEqual(schedule, expected_schedule)

    def test_scheduler(self):
        config = self._get_valid_config()
        expected_schedule = [
            0.1,
            0.1,
            0.1,
            0.1,
            0.01,
            0.01,
            0.001,
            0.001,
            0.0001,
            0.0001,
            0.0001,
            0.0001,
        ]
        self._test_config_scheduler(config, expected_schedule)

    def test_default_config(self):
        config = self._get_valid_config()
        default_config = copy.deepcopy(config)
        # Default equispaced drop_epochs behavior
        del default_config["milestones"]
        expected_schedule = [
            0.1,
            0.1,
            0.1,
            0.01,
            0.01,
            0.01,
            0.001,
            0.001,
            0.001,
            0.0001,
            0.0001,
            0.0001,
        ]
        self._test_config_scheduler(default_config, expected_schedule)

    def test_build_non_equi_step_scheduler(self):
        config = self._get_valid_config()
        scheduler = build_param_scheduler(config)
        self.assertTrue(isinstance(scheduler, MultiStepParamScheduler))



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from test.generic.config_utils import get_test_model_configs

from classy_vision.generic.profiler import (
    compute_activations,
    compute_flops,
    count_params,
)
from classy_vision.models import build_model


class TestProfilerFunctions(unittest.TestCase):
    def test_complexity_calculation(self) -> None:
        model_configs = get_test_model_configs()
        # make sure there are three configs returned
        self.assertEqual(len(model_configs), 3)

        # expected values which allow minor deviations from model changes
        # we only test at the 10^6 scale
        expected_m_flops = [4122, 7850, 8034]
        expected_m_params = [25, 44, 44]
        expected_m_activations = [11, 16, 21]

        for model_config, m_flops, m_params, m_activations in zip(
            model_configs, expected_m_flops, expected_m_params, expected_m_activations
        ):
            model = build_model(model_config)
            self.assertEqual(compute_activations(model) // 10 ** 6, m_activations)
            self.assertEqual(compute_flops(model) // 10 ** 6, m_flops)
            self.assertEqual(count_params(model) // 10 ** 6, m_params)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
This is the main script used for training Classy Vision jobs.

This can be used for training on your local machine, using CPU or GPU, and
for distributed training. This script also supports Tensorboard, Visdom and
checkpointing.

Example:
    For training locally, simply specify a configuration file and whether
    to use CPU or GPU:

        $ ./classy_train.py --device gpu --config configs/my_config.json

    For distributed training, this can be invoked via
    :func:`torch.distributed.launch`. For instance

        $ python -m torch.distributed.launch \
                 --nnodes=1 \
                 --nproc_per_node=1 \
                 --master_addr=localhost \
                 --master_port=29500 \
                 --use_env \
                 classy_train.py \
                 --device=gpu \
                 --config=configs/resnet50_synthetic_image_classy_config.json \
                 --num_workers=1 \
                 --log_freq=100

    For other use cases, try

        $ ./classy_train.py --help
"""

import logging
import os
from datetime import datetime
from pathlib import Path

import torch
from classy_vision.generic.opts import check_generic_args, parse_train_arguments
from classy_vision.generic.registry_utils import import_all_packages_from_directory
from classy_vision.generic.util import load_checkpoint, load_json
from classy_vision.hooks import (
    CheckpointHook,
    LossLrMeterLoggingHook,
    ProfilerHook,
    ProgressBarHook,
    TensorboardPlotHook,
    VisdomHook,
)
from classy_vision.tasks import FineTuningTask, build_task
from classy_vision.trainer import DistributedTrainer, LocalTrainer
from torchvision import set_image_backend, set_video_backend


try:
    import hydra

    hydra_available = True
except ImportError:
    hydra_available = False


def main(args, config):
    # Global flags
    torch.manual_seed(0)
    set_image_backend(args.image_backend)
    set_video_backend(args.video_backend)

    task = build_task(config)

    # Load checkpoint, if available.
    checkpoint = load_checkpoint(args.checkpoint_load_path)
    task.set_checkpoint(checkpoint)

    # Load a checkpoint contraining a pre-trained model. This is how we
    # implement fine-tuning of existing models.
    pretrained_checkpoint = load_checkpoint(args.pretrained_checkpoint_path)
    if pretrained_checkpoint is not None:
        assert isinstance(
            task, FineTuningTask
        ), "Can only use a pretrained checkpoint for fine tuning tasks"
        task.set_pretrained_checkpoint(pretrained_checkpoint)

    # Configure hooks to do tensorboard logging, checkpoints and so on
    task.set_hooks(configure_hooks(args, config))

    use_gpu = None
    if args.device is not None:
        use_gpu = args.device == "gpu"
        assert torch.cuda.is_available() or not use_gpu, "CUDA is unavailable"

    # LocalTrainer is used for a single node. DistributedTrainer will setup
    # training to use PyTorch's DistributedDataParallel.
    trainer_class = {"none": LocalTrainer, "ddp": DistributedTrainer}[
        args.distributed_backend
    ]

    trainer = trainer_class(use_gpu=use_gpu, num_dataloader_workers=args.num_workers)

    # That's it! When this call returns, training is done.
    trainer.train(task)

    output_folder = Path(args.checkpoint_folder).resolve()
    logging.info("Training successful!")
    logging.info(f'Results of this training run are available at: "{output_folder}"')


def configure_hooks(args, config):
    hooks = [LossLrMeterLoggingHook(args.log_freq)]

    # Make a folder to store checkpoints and tensorboard logging outputs
    suffix = datetime.now().isoformat()
    base_folder = f"{Path(__file__).parent}/output_{suffix}"
    if args.checkpoint_folder == "":
        args.checkpoint_folder = base_folder + "/checkpoints"
        os.makedirs(args.checkpoint_folder, exist_ok=True)

    logging.info(f"Logging outputs to {base_folder}")
    logging.info(f"Logging checkpoints to {args.checkpoint_folder}")

    if not args.skip_tensorboard:
        try:
            from tensorboardX import SummaryWriter

            tb_writer = SummaryWriter(log_dir=Path(base_folder) / "tensorboard")
            hooks.append(TensorboardPlotHook(tb_writer))
        except ImportError:
            logging.warning("tensorboardX not installed, skipping tensorboard hooks")

    args_dict = vars(args)
    args_dict["config"] = config
    hooks.append(
        CheckpointHook(
            args.checkpoint_folder, args_dict, checkpoint_period=args.checkpoint_period
        )
    )

    if args.profiler:
        hooks.append(ProfilerHook())
    if args.show_progress:
        hooks.append(ProgressBarHook())
    if args.visdom_server != "":
        hooks.append(VisdomHook(args.visdom_server, args.visdom_port))

    return hooks


if hydra_available:

    @hydra.main(config_path="hydra_configs/args.yaml")
    def hydra_main(cfg):
        args = cfg
        check_generic_args(cfg)
        config = cfg.config.to_container()
        main(args, config)


# run all the things:
if __name__ == "__main__":
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)

    logging.info("Classy Vision's default training script.")

    # This imports all modules in the same directory as classy_train.py
    # Because of the way Classy Vision's registration decorators work,
    # importing a module has a side effect of registering it with Classy
    # Vision. This means you can give classy_train.py a config referencing your
    # custom module (e.g. my_dataset) and it'll actually know how to
    # instantiate it.
    file_root = Path(__file__).parent
    import_all_packages_from_directory(file_root)

    if hydra_available:
        hydra_main()
    else:
        args = parse_train_arguments()
        config = load_json(args.config_file)
        main(args, config)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Callable, Dict, Optional, Sequence, Union

from classy_vision.dataset.transforms import ClassyTransform
from classy_vision.generic.distributed_util import get_rank, get_world_size
from classy_vision.generic.util import is_pos_int
from torch.utils.data import DataLoader
from torch.utils.data.distributed import DistributedSampler


def _return_true(_sample):
    return True


class ClassyDataset:
    """
    Class representing a dataset abstraction.

    This class wraps a :class:`torch.utils.data.Dataset` via the `dataset` attribute
    and configures the dataloaders needed to access the datasets.
    Transforms which need to be applied to the data should be specified in this class.
    ClassyDataset can be instantiated from a configuration file as well.
    """

    def __init__(
        self,
        dataset: Sequence,
        batchsize_per_replica: int,
        shuffle: bool,
        transform: Optional[Union[ClassyTransform, Callable]],
        num_samples: Optional[int],
    ) -> None:
        """
        Constructor for a ClassyDataset.

        Args:
            batchsize_per_replica: Positive integer indicating batch size for each
                replica
            shuffle: Whether to shuffle between epochs
            transform: When set, transform to be applied to each sample
            num_samples: When set, this restricts the number of samples provided by
                the dataset
        """
        # Asserts:
        assert is_pos_int(
            batchsize_per_replica
        ), "batchsize_per_replica must be a positive int"
        assert isinstance(shuffle, bool), "shuffle must be a boolean"
        assert num_samples is None or is_pos_int(
            num_samples
        ), "num_samples must be a positive int or None"

        # Assignments:
        self.batchsize_per_replica = batchsize_per_replica
        self.shuffle = shuffle
        self.transform = transform
        self.num_samples = num_samples
        self.dataset = dataset

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ClassyDataset":
        """Instantiates a ClassyDataset from a configuration.

        Args:
            config: A configuration for the ClassyDataset.

        Returns:
            A ClassyDataset instance.
        """
        raise NotImplementedError

    @classmethod
    def parse_config(cls, config: Dict[str, Any]):
        """
        This function parses out common config options.

        Args:
            config: A dict with the following string keys -

                | *batchsize_per_replica* (int): Must be a positive int, batch size
                |    for each replica
                | *use_shuffle* (bool): Whether to enable shuffling for the dataset
                | *num_samples* (int, optional): When set, restricts the number of
                     samples in a dataset
                | *transforms*: list of tranform configurations to be applied in order

        Returns:
            A tuple containing the following variables -
                | *transform_config*: Config for the dataset transform. Can be passed to
                |    :func:`transforms.build_transform`
                | *batchsize_per_replica*: Batch size per replica
                | *shuffle*: Whether we should shuffle between epochs
                | *num_samples*: When set, restricts the number of samples in a dataset
        """
        batchsize_per_replica = config.get("batchsize_per_replica")
        shuffle = config.get("use_shuffle")
        num_samples = config.get("num_samples")
        transform_config = config.get("transforms")
        return transform_config, batchsize_per_replica, shuffle, num_samples

    def __getitem__(self, idx: int):
        assert idx >= 0 and idx < len(
            self.dataset
        ), "Provided idx ({}) is outside of dataset range".format(idx)
        sample = self.dataset[idx]
        if self.transform is None:
            return sample
        return self.transform(sample)

    def __len__(self):
        assert self.num_samples is None or self.num_samples <= len(
            self.dataset
        ), "Num samples mus be less than length of base dataset"
        return len(self.dataset) if self.num_samples is None else self.num_samples

    def _get_sampler(self, epoch: int):
        """
        Return a :class:`torch.utils.data.sampler.Sampler` to sample the data.

        This is used to distribute the data across the replicas. If shuffling
        is enabled, every epoch will have a different shuffle.

        Args:
            epoch: The epoch being fetched.

        Returns:
            A sampler which tells the data loader which sample to load next.
        """
        world_size = get_world_size()
        rank = get_rank()
        sampler = DistributedSampler(
            self, num_replicas=world_size, rank=rank, shuffle=self.shuffle
        )
        sampler.set_epoch(epoch)
        return sampler

    def iterator(self, *args, **kwargs):
        """
        Returns an iterable which can be used to iterate over the data.

        Args:
            shuffle_seed (int, optional): Seed for the shuffle
            current_phase_id (int, optional): The epoch being fetched. Needed so that
                each epoch has a different shuffle order
        Returns:
            An iterable over the data
        """
        # TODO: Fix naming to be consistent (i.e. everyone uses epoch)
        shuffle_seed = kwargs.get("shuffle_seed", 0)
        assert isinstance(shuffle_seed, int), "Shuffle seed must be an int"
        epoch = kwargs.get("current_phase_id", 0)
        assert isinstance(epoch, int), "Epoch must be an int"

        offset_epoch = shuffle_seed + epoch

        return DataLoader(
            self,
            batch_size=self.batchsize_per_replica,
            num_workers=kwargs.get("num_workers", 0),
            pin_memory=kwargs.get("pin_memory", False),
            worker_init_fn=kwargs.get("worker_init_fn", None),
            multiprocessing_context=kwargs.get("multiprocessing_context", None),
            sampler=self._get_sampler(epoch=offset_epoch),
        )

    def get_batchsize_per_replica(self):
        """
        Get the batch size per replica.

        Returns:
            The batch size for each replica.
        """
        return self.batchsize_per_replica

    def get_global_batchsize(self):
        """
        Get the global batch size, combined over all the replicas.

        Returns:
            The overall batch size of the dataset.
        """
        return self.get_batchsize_per_replica() * get_world_size()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict

import numpy as np
from classy_vision.generic.util import convert_to_one_hot
from classy_vision.losses import ClassyLoss, register_loss
from classy_vision.losses.soft_target_cross_entropy_loss import (
    SoftTargetCrossEntropyLoss,
)


@register_loss("label_smoothing_cross_entropy")
class LabelSmoothingCrossEntropyLoss(ClassyLoss):
    def __init__(self, ignore_index, reduction, smoothing_param):
        """Intializer for the label smoothed cross entropy loss.
        This decreases gap between output scores and encourages generalization.
        Targets provided to forward can be one-hot vectors (NxC) or class indices(Nx1)

        Config params:
        'weight': weight of sample (not yet implemented),
        'ignore_index': sample should be ignored for loss (optional),
        'smoothing_param': value to be added to each target entry
        """
        super().__init__()
        self._ignore_index = ignore_index
        self._reduction = reduction
        self._smoothing_param = smoothing_param
        self.loss_function = SoftTargetCrossEntropyLoss(
            self._ignore_index, self._reduction, None
        )
        self._eps = np.finfo(np.float32).eps

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "LabelSmoothingCrossEntropyLoss":
        """Instantiates a LabelSmoothingCrossEntropyLoss from a configuration.

        Args:
            config: A configuration for a LabelSmoothingCrossEntropyLoss.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A LabelSmoothingCrossEntropyLoss instance.
        """

        assert "weight" not in config, '"weight" not implemented'
        assert (
            "smoothing_param" in config
        ), "Label Smoothing needs a smoothing parameter"
        return cls(
            ignore_index=config.get("ignore_index", -100),
            reduction=config.get("reduction", "mean"),
            smoothing_param=config.get("smoothing_param"),
        )

    def compute_valid_targets(self, target, classes):

        """
        This function takes one-hot or index target vectors and computes valid one-hot
        target vectors, based on ignore index value
        """
        target_shape_list = list(target.size())

        valid_mask = target != self._ignore_index
        valid_targets = target.float() * valid_mask.float()

        # check if targets are inputted as class integers
        if len(target_shape_list) == 1 or (
            len(target_shape_list) == 2 and target_shape_list[1] == 1
        ):

            valid_targets = convert_to_one_hot(valid_targets.view(-1, 1), classes)
            valid_targets = valid_targets.float() * valid_mask.view(-1, 1).float()

        return valid_targets

    def smooth_targets(self, valid_targets, classes):

        """
        This function takes valid (No ignore values present) one-hot target vectors
        and computes smoothed target vectors (normalized) according to the loss's
        smoothing parameter
        """

        valid_targets /= self._eps + valid_targets.sum(dim=1, keepdim=True)
        if classes > 0:
            smoothed_targets = valid_targets + (self._smoothing_param / classes)
        smoothed_targets /= self._eps + smoothed_targets.sum(dim=1, keepdim=True)

        return smoothed_targets

    def forward(self, output, target):
        valid_targets = self.compute_valid_targets(
            target=target, classes=output.shape[1]
        )
        assert (
            valid_targets.shape == output.shape
        ), "LabelSmoothingCrossEntropyLoss requires output and target to be same size"
        smoothed_targets = self.smooth_targets(
            valid_targets=valid_targets, classes=output.shape[1]
        )
        return self.loss_function(output, smoothed_targets)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import unittest

from classy_vision.optim.param_scheduler import build_param_scheduler
from classy_vision.optim.param_scheduler.constant_scheduler import (
    ConstantParamScheduler,
)


class TestFixedScheduler(unittest.TestCase):
    _num_epochs = 12

    def _get_valid_config(self):
        return {"name": "constant", "num_epochs": self._num_epochs, "value": 0.1}

    def test_invalid_config(self):
        # Invalid num epochs
        config = self._get_valid_config()

        bad_config = copy.deepcopy(config)
        del bad_config["value"]
        with self.assertRaises(AssertionError):
            ConstantParamScheduler.from_config(bad_config)

    def test_scheduler(self):
        config = self._get_valid_config()

        scheduler = ConstantParamScheduler.from_config(config)
        schedule = [
            scheduler(epoch_num / self._num_epochs)
            for epoch_num in range(self._num_epochs)
        ]
        expected_schedule = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1]

        self.assertEqual(schedule, expected_schedule)
        # The input for the scheduler should be in the interval [0;1), open
        with self.assertRaises(RuntimeError):
            scheduler(1)

    def test_build_constant_scheduler(self):
        config = self._get_valid_config()
        scheduler = build_param_scheduler(config)
        self.assertTrue(isinstance(scheduler, ConstantParamScheduler))



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


class MergeDataset:
    """
        Dataset that merges samples from multiple datasets into single sample.

        If datasets have distinct keys, then we merge dicts, e.g.

            dataset1[idx] = {'input': input_tensor}
            dataset2[idx] = {'target': target_tensor}
            merged_dataset[idx] = {'input': input_tensor, 'target': target_tensor}

        If datasets have matching keys then we create a list and append, e.g.

            dataset1[idx] = {'input': input_tensor1}
            dataset2[idx] = {'input': input_tensor2}
            merged_dataset[idx] = {'input': [input_tensor1, input_tensor2]}

        Note, if your datasets' samples do not have consistent keys for each sample,
        this could lead to inconsistent samples merged samples.
    """

    def __init__(self, datasets):

        # assertions:
        assert isinstance(datasets, list)
        assert all(len(dataset) == len(datasets[0]) for dataset in datasets)

        # create object:
        super(MergeDataset, self).__init__()
        self.datasets = datasets

    def __getitem__(self, idx):
        final_sample = {}
        for dataset in self.datasets:
            curr_sample = dataset[idx]
            assert isinstance(curr_sample, dict), "Merge dataset only supports dicts"
            for key in curr_sample.keys():
                # If keys are distinct, then
                if key not in final_sample:
                    final_sample[key] = curr_sample[key]
                elif not isinstance(final_sample[key], list):
                    final_sample[key] = [final_sample[key], curr_sample[key]]
                else:
                    final_sample[key].append(curr_sample[key])

        return final_sample

    def __len__(self):
        return len(self.datasets[0])



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Any, Callable, Dict, List, Optional

import torchvision.transforms as transforms

from . import register_dataset
from .classy_dataset import ClassyDataset
from .core import RandomImageBinaryClassDataset
from .transforms import build_transforms
from .transforms.util import ImagenetConstants, build_field_transform_default_imagenet


@register_dataset("synthetic_image")
class SyntheticImageDataset(ClassyDataset):
    """Classy Dataset which produces random synthetic images with binary targets.

    The underlying dataset sets targets based on the channels in the image, so users can
    validate their setup by checking if they can get 100% accuracy on this dataset.
    Useful for testing since the dataset is much faster to initialize and fetch samples
    from, compared to real world datasets.
    """

    def __init__(
        self,
        batchsize_per_replica: int,
        shuffle: bool,
        transform: Optional[Callable],
        num_samples: int,
        crop_size: int,
        class_ratio: float,
        seed: int,
    ) -> None:
        """
        Args:
            batchsize_per_replica: Positive integer indicating batch size for each
                replica
            shuffle: Whether we should shuffle between epochs
            transform: When specified, transform to be applied to each sample
            num_samples: Number of samples to return
            crop_size: Image size, used for both height and width
            class_ratio: Ratio of the distribution of target classes
            seed: Seed used for image generation. Use the same seed to generate the same
                set of samples.
            split: When specified, split of dataset to use
        """
        dataset = RandomImageBinaryClassDataset(
            crop_size, class_ratio, num_samples, seed
        )
        super().__init__(
            dataset, batchsize_per_replica, shuffle, transform, num_samples
        )

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "SyntheticImageDataset":
        """Instantiates a SyntheticImageDataset from a configuration.

        Args:
            config: A configuration for a SyntheticImageDataset.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A SyntheticImageDataset instance.
        """
        assert all(key in config for key in ["crop_size", "class_ratio", "seed"])
        crop_size = config["crop_size"]
        class_ratio = config["class_ratio"]
        seed = config["seed"]
        (
            transform_config,
            batchsize_per_replica,
            shuffle,
            num_samples,
        ) = cls.parse_config(config)

        try:
            transform = build_transforms(transform_config)
        except Exception:
            logging.error(
                "We recently changed transform behavior"
                " do you need to update your config?"
                " See resnet50_synthetic_image_classy_config.json"
                " as an example."
            )
            raise

        return cls(
            batchsize_per_replica,
            shuffle,
            transform,
            num_samples,
            crop_size,
            class_ratio,
            seed,
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import warnings
from pathlib import Path

from classy_vision.generic.registry_utils import import_all_modules

from .classy_optimizer import ClassyOptimizer
from .param_scheduler import build_param_scheduler


FILE_ROOT = Path(__file__).parent


OPTIMIZER_REGISTRY = {}
OPTIMIZER_CLASS_NAMES = set()


def build_optimizer(config):
    """Builds a ClassyOptimizer from a config.

    This assumes a 'name' key in the config which is used to determine what
    optimizer class to instantiate. For instance, a config `{"name": "my_optimizer",
    "foo": "bar"}` will find a class that was registered as "my_optimizer"
    (see :func:`register_optimizer`) and call .from_config on it.

    Also builds the param schedulers passed in the config and associates them with the
    optimizer. The config should contain an optional "param_schedulers" key containing a
    dictionary of param scheduler configs, keyed by the parameter they control. Adds
    "num_epochs" to each of the scheduler configs and then calls
    :func:`build_param_scheduler` on each config in the dictionary.
    """
    optimizer = OPTIMIZER_REGISTRY[config["name"]].from_config(config)

    # create a deepcopy since we will be modifying the param scheduler config
    param_scheduler_config = copy.deepcopy(config.get("param_schedulers", {}))

    # build the param schedulers
    if "lr" in config and isinstance(config["lr"], dict):
        message = (
            'Passing an lr schedule in the config using "lr" is deprecated and '
            "will be removed in version 0.2.0. See the docs for build_optimizer for "
            "the recommended format."
        )
        warnings.warn(message, DeprecationWarning, stacklevel=2)
        assert (
            param_scheduler_config == {}
        ), 'Cannot pass both "lr" and "param_schedulers" to the config'
        param_scheduler_config = {"lr": config["lr"]}
    for cfg in param_scheduler_config.values():
        cfg["num_epochs"] = config["num_epochs"]

    param_schedulers = {
        param: build_param_scheduler(cfg)
        for param, cfg in param_scheduler_config.items()
    }
    optimizer.set_param_schedulers(param_schedulers)
    return optimizer


def register_optimizer(name):
    """Registers a ClassyOptimizer subclass.

    This decorator allows Classy Vision to instantiate a subclass of
    ClassyOptimizer from a configuration file, even if the class itself is not
    part of the Classy Vision framework. To use it, apply this decorator to a
    ClassyOptimizer subclass, like this:

    .. code-block:: python

        @register_optimizer('my_optimizer')
        class MyOptimizer(ClassyOptimizer):
            ...

    To instantiate an optimizer from a configuration file, see
    :func:`build_optimizer`."""

    def register_optimizer_cls(cls):
        if name in OPTIMIZER_REGISTRY:
            raise ValueError("Cannot register duplicate optimizer ({})".format(name))
        if not issubclass(cls, ClassyOptimizer):
            raise ValueError(
                "Optimizer ({}: {}) must extend ClassyVisionOptimizer".format(
                    name, cls.__name__
                )
            )
        if cls.__name__ in OPTIMIZER_CLASS_NAMES:
            raise ValueError(
                "Cannot register optimizer with duplicate class name({})".format(
                    cls.__name__
                )
            )
        OPTIMIZER_REGISTRY[name] = cls
        OPTIMIZER_CLASS_NAMES.add(cls.__name__)
        return cls

    return register_optimizer_cls


# automatically import any Python files in the optim/ directory
import_all_modules(FILE_ROOT, "classy_vision.optim")

from .adam import Adam  # isort:skip
from .rmsprop import RMSProp  # isort:skip
from .rmsprop_tf import RMSPropTF  # isort:skip
from .sgd import SGD  # isort:skip

__all__ = [
    "Adam",
    "ClassyOptimizer",
    "RMSProp",
    "RMSPropTF",
    "SGD",
    "build_optimizer",
    "register_optimizer",
]



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import unittest
from test.generic.utils import compare_model_state

import torch
from classy_vision.models import ClassyModel, build_model


class TestResNeXt3D(unittest.TestCase):
    def setUp(self):
        model_config_template = {
            "name": "resnext3d",
            "input_key": "video",
            "clip_crop_size": 112,
            "skip_transformation_type": "postactivated_shortcut",
            "frames_per_clip": 32,
            "input_planes": 3,
            "stem_name": "resnext3d_stem",
            "stem_planes": 64,
            "stem_temporal_kernel": 3,
            "stage_planes": 64,
            "num_groups": 1,
            "width_per_group": 16,
            "heads": [
                {
                    "name": "fully_convolutional_linear",
                    "unique_id": "default_head",
                    "in_plane": 512,
                    "pool_size": (2, 7, 7),
                    "activation_func": "softmax",
                    "num_classes": 2,
                }
            ],
        }
        pbt = "postactivated_bottleneck_transformation"
        model_config_variants = [
            # ResNeXt3D-34
            {
                "residual_transformation_type": "basic_transformation",
                "num_blocks": [3, 4, 6, 3],
            },
            # ResNeXt3D-50
            {"residual_transformation_type": pbt, "num_blocks": [3, 4, 6, 3]},
            # ResNeXt3D-101
            {"residual_transformation_type": pbt, "num_blocks": [3, 4, 23, 3]},
        ]

        self.model_configs = []
        for variant in model_config_variants:
            model_config = copy.deepcopy(model_config_template)
            model_config.update(variant)

            block_idx = model_config["num_blocks"][-1]
            # attach the head at the last block
            model_config["heads"][0]["fork_block"] = "pathway0-stage4-block%d" % (
                block_idx - 1
            )

            self.model_configs.append(model_config)

        self.batchsize = 1

        self.forward_pass_configs = {
            "train": {
                # input shape: N x C x T x H x W
                "input": {"video": torch.rand(self.batchsize, 3, 16, 112, 112)},
                "model": {
                    "stem_maxpool": False,
                    "stage_temporal_stride": [1, 2, 2, 2],
                    "stage_spatial_stride": [1, 2, 2, 2],
                },
            },
            "test": {
                "input": {"video": torch.rand(self.batchsize, 3, 16, 256, 320)},
                "model": {
                    "stem_maxpool": True,
                    "stage_temporal_stride": [1, 2, 2, 2],
                    "stage_spatial_stride": [1, 2, 2, 2],
                },
            },
        }

    def test_build_model(self):
        for model_config in self.model_configs:
            model = build_model(model_config)
            self.assertTrue(isinstance(model, ClassyModel))
            self.assertTrue(
                type(model.output_shape) == tuple and len(model.output_shape) == 2
            )
            self.assertTrue(type(model.model_depth) == int)

    def test_forward_pass(self):
        for split, split_config in self.forward_pass_configs.items():
            for model_config in self.model_configs:
                forward_pass_model_config = copy.deepcopy(model_config)
                forward_pass_model_config.update(split_config["model"])

                num_classes = forward_pass_model_config["heads"][0]["num_classes"]

                model = build_model(forward_pass_model_config)
                model.train(split == "train")

                out = model(split_config["input"])

                self.assertEqual(out.size(), (self.batchsize, num_classes))

    def test_set_classy_state_plain(self):
        # We use the same model architecture to save and load a model state.
        # This is a plain use case of `set_classy_state` method
        for model_config in self.model_configs:
            model = build_model(model_config)
            model_state = model.get_classy_state()

            model2 = build_model(model_config)
            model2.set_classy_state(model_state)
            model2_state = model2.get_classy_state()
            compare_model_state(self, model_state, model2_state)

    def _get_model_config_weight_inflation(self):
        model_2d_config = {
            "name": "resnext3d",
            "frames_per_clip": 1,
            "input_planes": 3,
            "clip_crop_size": 224,
            "skip_transformation_type": "postactivated_shortcut",
            "residual_transformation_type": "postactivated_bottleneck_transformation",
            "num_blocks": [3, 4, 6, 3],
            "stem_name": "resnext3d_stem",
            "stem_planes": 64,
            "stem_temporal_kernel": 1,
            "stem_spatial_kernel": 7,
            "stem_maxpool": True,
            "stage_planes": 256,
            "stage_temporal_kernel_basis": [[1], [1], [1], [1]],
            "temporal_conv_1x1": [True, True, True, True],
            "stage_temporal_stride": [1, 1, 1, 1],
            "stage_spatial_stride": [1, 2, 2, 2],
            "num_groups": 1,
            "width_per_group": 64,
            "num_classes": 1000,
            "zero_init_residual_transform": True,
            "heads": [
                {
                    "name": "fully_convolutional_linear",
                    "unique_id": "default_head",
                    "pool_size": [1, 7, 7],
                    "activation_func": "softmax",
                    "num_classes": 1000,
                    "fork_block": "pathway0-stage4-block2",
                    "in_plane": 2048,
                    "use_dropout": False,
                }
            ],
        }

        model_3d_config = {
            "name": "resnext3d",
            "frames_per_clip": 8,
            "input_planes": 3,
            "clip_crop_size": 224,
            "skip_transformation_type": "postactivated_shortcut",
            "residual_transformation_type": "postactivated_bottleneck_transformation",
            "num_blocks": [3, 4, 6, 3],
            "input_key": "video",
            "stem_name": "resnext3d_stem",
            "stem_planes": 64,
            "stem_temporal_kernel": 5,
            "stem_spatial_kernel": 7,
            "stem_maxpool": True,
            "stage_planes": 256,
            "stage_temporal_kernel_basis": [[3], [3, 1], [3, 1], [1, 3]],
            "temporal_conv_1x1": [True, True, True, True],
            "stage_temporal_stride": [1, 1, 1, 1],
            "stage_spatial_stride": [1, 2, 2, 2],
            "num_groups": 1,
            "width_per_group": 64,
            "num_classes": 1000,
            "freeze_trunk": False,
            "zero_init_residual_transform": True,
            "heads": [
                {
                    "name": "fully_convolutional_linear",
                    "unique_id": "default_head",
                    "pool_size": [8, 7, 7],
                    "activation_func": "softmax",
                    "num_classes": 1000,
                    "fork_block": "pathway0-stage4-block2",
                    "in_plane": 2048,
                    "use_dropout": True,
                }
            ],
        }
        return model_2d_config, model_3d_config

    def test_set_classy_state_weight_inflation(self):
        # Get model state from a 2D ResNet model, inflate the 2D conv weights,
        # and use them to initialize 3D conv weights. This is an advanced use of
        # `set_classy_state` method.
        model_2d_config, model_3d_config = self._get_model_config_weight_inflation()
        model_2d = build_model(model_2d_config)
        model_2d_state = model_2d.get_classy_state()

        model_3d = build_model(model_3d_config)
        model_3d.set_classy_state(model_2d_state)
        model_3d_state = model_3d.get_classy_state()

        for name, weight_2d in model_2d_state["model"]["trunk"].items():
            weight_3d = model_3d_state["model"]["trunk"][name]
            if weight_2d.dim() == 5:
                # inflation only applies to conv weights
                self.assertEqual(weight_3d.dim(), 5)
                if weight_2d.shape[2] == 1 and weight_3d.shape[2] > 1:
                    weight_2d_inflated = (
                        weight_2d.repeat(1, 1, weight_3d.shape[2], 1, 1)
                        / weight_3d.shape[2]
                    )
                    self.assertTrue(torch.equal(weight_3d, weight_2d_inflated))

    def test_set_classy_state_weight_inflation_inconsistent_kernel_size(self):
        # Get model state from a 2D ResNet model, inflate the 2D conv weights,
        # and use them to initialize 3D conv weights.
        model_2d_config, model_3d_config = self._get_model_config_weight_inflation()
        # Modify conv kernel size in the stem layer of 2D model to 5, which is
        # inconsistent with the kernel size 7 used in 3D model.
        model_2d_config["stem_spatial_kernel"] = 5
        model_2d = build_model(model_2d_config)
        model_2d_state = model_2d.get_classy_state()
        model_3d = build_model(model_3d_config)
        with self.assertRaises(AssertionError):
            model_3d.set_classy_state(model_2d_state)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest

import torch
import torchvision.transforms as transforms
from classy_vision.dataset.transforms import (
    ClassyTransform,
    build_transforms,
    register_transform,
)
from classy_vision.dataset.transforms.util import ImagenetNoAugmentTransform


@register_transform("resize")
class resize(ClassyTransform):
    def __init__(self, size: int):
        self.transform = transforms.Resize(size=size)

    def __call__(self, img):
        return self.transform(img)


@register_transform("center_crop")
class center_crop(ClassyTransform):
    def __init__(self, size: int):
        self.transform = transforms.CenterCrop(size=size)

    def __call__(self, img):
        return self.transform(img)


class DatasetTransformsTest(unittest.TestCase):
    def get_test_image(self):
        return transforms.ToPILImage()(torch.randn((3, 224, 224)))

    def test_transforms(self):
        input = self.get_test_image()

        # reference transform which we will use to validate the built transforms
        reference_transform = ImagenetNoAugmentTransform()
        reference_output = reference_transform(input)

        # test a registered transform
        config = [{"name": "imagenet_no_augment"}]
        transform = build_transforms(config)
        output = transform(input)
        self.assertTrue(torch.allclose(output, reference_output))

        # test a transform built using torchvision transforms
        config = [
            {"name": "Resize", "size": 256},
            {"name": "CenterCrop", "size": 224},
            {"name": "ToTensor"},
            {
                "name": "Normalize",
                "mean": [0.485, 0.456, 0.406],
                "std": [0.229, 0.224, 0.225],
            },
        ]
        transform = build_transforms(config)
        output = transform(input)
        self.assertTrue(torch.allclose(output, reference_output))

        # test a combination of registered and torchvision transforms
        config = [
            {"name": "resize", "size": 256},
            {"name": "center_crop", "size": 224},
            {"name": "ToTensor"},
            {
                "name": "Normalize",
                "mean": [0.485, 0.456, 0.406],
                "std": [0.229, 0.224, 0.225],
            },
        ]
        transform = build_transforms(config)
        output = transform(input)
        self.assertTrue(torch.allclose(output, reference_output))



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch.nn.functional as F
from classy_vision.losses import ClassyLoss, register_loss


@register_loss("my_loss")
class MyLoss(ClassyLoss):
    def forward(self, input, target):
        labels = F.one_hot(target, num_classes=2).float()
        return F.binary_cross_entropy(input, labels)

    @classmethod
    def from_config(cls, config):
        # We don't need anything from the config
        return cls()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
from typing import Any, Dict

from classy_vision import tasks
from classy_vision.generic.profiler import (
    compute_activations,
    compute_flops,
    count_params,
)
from classy_vision.hooks.classy_hook import ClassyHook


class ModelComplexityHook(ClassyHook):
    """
    Logs the number of paramaters and forward pass FLOPs and activations of the model.
    """

    on_phase_start = ClassyHook._noop
    on_forward = ClassyHook._noop
    on_loss_and_meter = ClassyHook._noop
    on_step = ClassyHook._noop
    on_phase_end = ClassyHook._noop
    on_end = ClassyHook._noop

    def on_start(
        self, task: "tasks.ClassyTask", local_variables: Dict[str, Any]
    ) -> None:
        """Measure number of parameters, FLOPs and activations."""
        self.num_flops = 0
        self.num_activations = 0
        self.num_parameters = 0
        try:
            self.num_parameters = count_params(task.base_model)
            logging.info("Number of parameters in model: %d" % self.num_parameters)
            try:
                self.num_flops = compute_flops(
                    task.base_model,
                    input_shape=task.base_model.input_shape,
                    input_key=task.base_model.input_key
                    if hasattr(task.base_model, "input_key")
                    else None,
                )
                if self.num_flops is None:
                    logging.info("FLOPs for forward pass: skipped.")
                    self.num_flops = 0
                else:
                    logging.info(
                        "FLOPs for forward pass: %d MFLOPs"
                        % (float(self.num_flops) / 1e6)
                    )
            except NotImplementedError:
                logging.warning(
                    """Model contains unsupported modules:
                Could not compute FLOPs for model forward pass. Exception:""",
                    exc_info=True,
                )
            try:
                self.num_activations = compute_activations(
                    task.base_model,
                    input_shape=task.base_model.input_shape,
                    input_key=task.base_model.input_key
                    if hasattr(task.base_model, "input_key")
                    else None,
                )
                logging.info(f"Number of activations in model: {self.num_activations}")
            except NotImplementedError:
                logging.info(
                    "Model does not implement input_shape. Skipping "
                    "activation calculation."
                )
        except Exception:
            logging.exception("Unexpected failure estimating model complexity.")

    def get_summary(self):
        return {
            "FLOPS(M)": float(self.num_flops) / 1e6
            if self.num_flops is not None
            else 0,
            "num_activations(M)": float(self.num_activations) / 1e6
            if self.num_activations is not None
            else 0,
            "num_parameters(M)": float(self.num_parameters) / 1e6
            if self.num_parameters is not None
            else 0,
        }



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from pathlib import Path
from typing import Any, Dict

from classy_vision.generic.registry_utils import import_all_modules

from .classy_vision_param_scheduler import (  # noqa F401
    ClassyParamScheduler,
    UpdateInterval,
)


FILE_ROOT = Path(__file__).parent


PARAM_SCHEDULER_REGISTRY = {}


def build_param_scheduler(config: Dict[str, Any]) -> ClassyParamScheduler:
    """Builds a :class:`ClassyParamScheduler` from a config.

    This assumes a 'name' key in the config which is used to determine what
    param scheduler class to instantiate. For instance, a config `{"name":
    "my_scheduler", "foo": "bar"}` will find a class that was registered as
    "my_scheduler" (see :func:`register_param_scheduler`) and call .from_config
    on it."""
    return PARAM_SCHEDULER_REGISTRY[config["name"]].from_config(config)


def register_param_scheduler(name):
    """Registers a :class:`ClassyParamScheduler` subclass.

    This decorator allows Classy Vision to instantiate a subclass of
    ClassyParamScheduler from a configuration file, even if the class itself is not
    part of the Classy Vision framework. To use it, apply this decorator to a
    ClassyParamScheduler subclass, like this:

    .. code-block:: python

        @register_param_scheduler('my_scheduler')
        class MyParamScheduler(ClassyParamScheduler):
            ...

    To instantiate a param scheduler from a configuration file, see
    :func:`build_param_scheduler`."""

    def register_param_scheduler_cls(cls):
        if name in PARAM_SCHEDULER_REGISTRY:
            raise ValueError(
                "Cannot register duplicate param scheduler ({})".format(name)
            )
        if not issubclass(cls, ClassyParamScheduler):
            raise ValueError(
                "Param Scheduler ({}: {}) must extend ClassyParamScheduler".format(
                    name, cls.__name__
                )
            )
        PARAM_SCHEDULER_REGISTRY[name] = cls
        return cls

    return register_param_scheduler_cls


# automatically import any Python files in the optim/param_scheduler/ directory
import_all_modules(FILE_ROOT, "classy_vision.optim.param_scheduler")

from .composite_scheduler import CompositeParamScheduler, IntervalScaling  # isort:skip
from .constant_scheduler import ConstantParamScheduler  # isort:skip
from .cosine_scheduler import CosineParamScheduler  # isort:skip
from .linear_scheduler import LinearParamScheduler  # isort:skip
from .multi_step_scheduler import MultiStepParamScheduler  # isort:skip
from .polynomial_decay_scheduler import PolynomialDecayParamScheduler  # isort:skip
from .step_scheduler import StepParamScheduler  # isort:skip
from .step_with_fixed_gamma_scheduler import (  # isort:skip
    StepWithFixedGammaParamScheduler,
)

__all__ = [
    "ClassyParamScheduler",
    "CompositeParamScheduler",
    "ConstantParamScheduler",
    "CosineParamScheduler",
    "LinearParamScheduler",
    "MultiStepParamScheduler",
    "PolynomialDecayParamScheduler",
    "StepParamScheduler",
    "UpdateInterval",
    "IntervalScaling",
    "StepWithFixedGammaParamScheduler",
    "build_param_scheduler",
    "register_param_scheduler",
]



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import unittest

import torch
import torchvision.transforms._transforms_video as transforms_video
from classy_vision.dataset.core.random_video_datasets import RandomVideoDataset
from classy_vision.dataset.transforms.util_video import (
    VideoConstants,
    build_video_field_transform_default,
)


class DatasetTransformUtilVideoTest(unittest.TestCase):
    def get_test_video_dataset(self):
        self.num_classes = 10
        self.split = "train"
        self.num_samples = 100
        self.frames_per_clip = 32
        self.video_width = 320
        self.video_height = 256
        self.audio_samples = 44000
        self.clips_per_video = 1
        self.seed = 1

        dataset = RandomVideoDataset(
            self.num_classes,
            self.split,
            self.num_samples,
            self.frames_per_clip,
            self.video_width,
            self.video_height,
            self.audio_samples,
            self.clips_per_video,
            self.seed,
        )
        return dataset

    def test_build_field_transform_default_video(self):
        dataset = self.get_test_video_dataset()

        # transform config is not provided. Use default transforms
        config = None
        # default training data transform
        sample = dataset[0]

        transform = build_video_field_transform_default(config, "train")
        output_clip = transform(sample)["input"]["video"]
        self.assertEqual(
            output_clip.size(),
            torch.Size(
                (
                    3,
                    self.frames_per_clip,
                    VideoConstants.CROP_SIZE,
                    VideoConstants.CROP_SIZE,
                )
            ),
        )
        # default testing data transform
        sample = dataset[1]
        sample_copy = copy.deepcopy(sample)

        expected_output_clip = transforms_video.ToTensorVideo()(
            sample["input"]["video"]
        )
        expected_output_clip = transforms_video.CenterCropVideo(
            VideoConstants.CROP_SIZE
        )(expected_output_clip)
        expected_output_clip = transforms_video.NormalizeVideo(
            mean=VideoConstants.MEAN, std=VideoConstants.STD
        )(expected_output_clip)

        transform = build_video_field_transform_default(config, "test")
        output_clip = transform(sample_copy)["input"]["video"]

        rescaled_width = int(
            VideoConstants.SIZE_RANGE[0] * self.video_width / self.video_height
        )
        self.assertEqual(
            output_clip.size(),
            torch.Size(
                (3, self.frames_per_clip, VideoConstants.SIZE_RANGE[0], rescaled_width)
            ),
        )
        # transform config is provided. Simulate training config
        sample = dataset[2]
        config = {
            "video": [
                {"name": "ToTensorVideo"},
                {
                    "name": "video_clip_random_resize_crop",
                    "crop_size": 64,
                    "size_range": [256, 320],
                },
                {"name": "RandomHorizontalFlipVideo"},
                {
                    "name": "NormalizeVideo",
                    "mean": [0.485, 0.456, 0.406],
                    "std": [0.229, 0.224, 0.225],
                },
            ]
        }
        transform = build_video_field_transform_default(config, "train")
        output_clip = transform(sample)["input"]["video"]
        self.assertEqual(
            output_clip.size(), torch.Size((3, self.frames_per_clip, 64, 64))
        )
        self.assertTrue(output_clip.dtype == torch.float)

        # transform config is provided. Simulate testing config
        sample = dataset[3]
        config = {
            "video": [
                {"name": "ToTensorVideo"},
                {"name": "video_clip_resize", "size": 64},
                {
                    "name": "NormalizeVideo",
                    "mean": [0.485, 0.456, 0.406],
                    "std": [0.229, 0.224, 0.225],
                },
            ]
        }
        transform = build_video_field_transform_default(config, "train")
        output_clip = transform(sample)["input"]["video"]

        rescaled_width = int(64 * self.video_width / self.video_height)
        self.assertEqual(
            output_clip.size(),
            torch.Size((3, self.frames_per_clip, 64, rescaled_width)),
        )
        self.assertTrue(output_clip.dtype == torch.float)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict

from . import ClassyParamScheduler, register_param_scheduler


@register_param_scheduler("polynomial")
class PolynomialDecayParamScheduler(ClassyParamScheduler):
    """
    Decays the param value after every epoch according to a
    polynomial function with a fixed power.

    Example:

        .. code-block:: python

          base_lr: 0.1
          power: 0.9

    Then the param value will be 0.1 for epoch 0, 0.099 for epoch 1, and
    so on.
    """

    def __init__(self, base_lr, power):
        super().__init__()

        self._base_lr = base_lr
        self._power = power

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "PolynomialDecayParamScheduler":
        """Instantiates a PolynomialDecayParamScheduler from a configuration.

        Args:
            config: A configuration for a PolynomialDecayParamScheduler.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A PolynomialDecayParamScheduler instance.
        """
        assert (
            "base_lr" in config and "power" in config
        ), "Polynomial decay scheduler requires a base lr and a power of decay"
        return cls(base_lr=config["base_lr"], power=config["power"])

    def __call__(self, where: float):
        return self._base_lr * (1 - where) ** self._power



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import enum
import logging
import time
from typing import Any, Dict, List, Optional, Union

import torch
from classy_vision.dataset import ClassyDataset, build_dataset
from classy_vision.generic.distributed_util import (
    all_reduce_mean,
    barrier,
    init_distributed_data_parallel_model,
    is_distributed_training_run,
)
from classy_vision.generic.util import (
    copy_model_to_gpu,
    recursive_copy_to_gpu,
    update_classy_state,
)
from classy_vision.hooks import ClassyHookFunctions
from classy_vision.losses import ClassyLoss, build_loss
from classy_vision.meters import build_meters
from classy_vision.models import ClassyModel, build_model
from classy_vision.optim import ClassyOptimizer, build_optimizer
from torch.distributed import broadcast

from . import register_task
from .classy_task import ClassyTask


try:
    import apex

    apex_available = True
except ImportError:
    apex_available = False


class BroadcastBuffersMode(enum.Enum):
    DISABLED = enum.auto()
    # Enable DistributedDataParallel's broadcast_buffers option, synchronizing
    # model buffers every forward pass.
    FORWARD_PASS = enum.auto()
    # Similar to FORWARD_PASS, but only synchronizes model buffers once
    # per epoch, between train and test phases. If your motivation for
    # synchronizing buffers is for buffers to be consistent during eval, use
    # this instead of FORWARD_PASS to reduce training overhead.
    BEFORE_EVAL = enum.auto()


@register_task("classification_task")
class ClassificationTask(ClassyTask):
    """Basic classification training task.

    This task encapsultates all of the components and steps needed to
    train a classifier using a :class:`classy_vision.trainer.ClassyTrainer`.

    Assumes a train / test phase per each epoch and that the datasets
    have the same API as the map-style Dataset class in
    `torch.utils.data.dataset <https://pytorch.org/docs/stable/data.html
    #torch.utils.data.Dataset>`_ (in particular, this task makes use of
    the len).  If you are using an `IterableDataset <https://pytorch.org/docs/
    stable/data.html#torch.utils.data.IterableDataset>`_ then a custom task
    may be appropriate.


    :var loss: Loss (see :class:`classy_vision.losses.ClassyLoss`) function used
        for computing the loss in each forward pass
    :var datasets: Mapping from a ``phase_type`` in ["train", "test']
        to dataset used for training (or testing)
    :var meters: List of meters (see :class:`classy_vision.meters.ClassyMeter`)
        to calculate during training
    :var num_epochs: Number of epochs (passes over dataset) to train
    :var test_only: Used to only run the test phase
    :var base_model: Model to be trained, unwrapped in DDP or DP wrappers
    :var optimizer: Optimizer used in train step
    :var checkpoint: Serializable dict which represents state in training
    :var phases: List of phase specific information, e.g. if phase is
        train / test.
    :var hooks: List of hooks to apply during training
    :var train: Phase type, if true it means we are training,
        false means testing
    :var distributed_model: Base model, but wrapped in DDP (DistributedDataParallel)
    :var phase_idx: Current phase id, first phase is 0, if task has not started
        training then returns -1
    :var train_phase_idx: Only counts train phases
    :var num_updates: Number of total parameter updates applied to model
        by the optimizer
    :var data_iterator: Iterator which can be used to obtain batches
    :var losses: Loss curve
    :var perf_log: list of training speed measurements, to be logged

    """

    def __init__(self):
        """Constructs a ClassificationTask
        """
        super().__init__()

        self.loss = None
        self.datasets = {}
        self.meters = []
        self.num_epochs = 1
        self.test_phase_period = 1
        self.test_only = False
        self.base_model = None
        self.optimizer = None
        self.checkpoint = None
        self.phases = []
        self.hooks = []
        self.train = True
        self.distributed_model = None
        self.phase_idx = -1
        self.train_phase_idx = -1
        self.num_updates = 0
        self.data_iterator = None
        self.losses = []
        self.broadcast_buffers_mode: BroadcastBuffersMode = (
            BroadcastBuffersMode.DISABLED
        )
        self.amp_opt_level = None
        self.perf_log = []

    def set_checkpoint(self, checkpoint):
        """Sets checkpoint on task.

        Args:
            checkpoint: A serializable dict representing current task state
        """
        assert (
            checkpoint is None or "classy_state_dict" in checkpoint
        ), "Checkpoint does not contain classy_state_dict"
        self.checkpoint = checkpoint

    def set_num_epochs(self, num_epochs: Union[int, float]):
        """Set number of epochs to be run.

        Args:
           num_epochs: Number of epochs to run task
        """
        self.num_epochs = num_epochs
        return self

    def set_test_phase_period(self, test_phase_period: int):
        """Set the period of test phase.

        Args:
            test_phase_period: The period of test phase
        """
        self.test_phase_period = test_phase_period
        return self

    def set_dataset(self, dataset: ClassyDataset, phase_type: str):
        """Set dataset for phase type on task

        Args:
            dataset: ClassyDataset for returning samples.
            phase_type: str must be one of "train" or "test"
        """
        assert phase_type in [
            "train",
            "test",
        ], "phase_type must be in ['train', 'test']"
        self.datasets[phase_type] = dataset
        return self

    def set_optimizer(self, optimizer: ClassyOptimizer):
        """Set optimizer for task

        Args:
            optimizer: optimizer for task
        """
        self.optimizer = optimizer
        return self

    def set_loss(self, loss: ClassyLoss):
        """Set loss function for task

        Args:
            loss: loss for task
        """
        self.loss = loss
        return self

    def set_meters(self, meters: List["ClassyMeter"]):
        """Set meters for task

        Args:
            meters: list of meters to compute during training
        """
        self.meters = meters
        return self

    def set_distributed_options(self, broadcast_buffers_mode: BroadcastBuffersMode):
        """Set distributed options.

        Args:
            broadcast_buffers_mode: Broadcast buffers mode. See
                :class:`BroadcastBuffersMode` for options.
        """
        self.broadcast_buffers_mode = broadcast_buffers_mode
        return self

    def set_hooks(self, hooks: List["ClassyHook"]):
        """Set hooks for task

        Args:
            hooks: List of hooks to apply during training
        """
        from classy_vision.hooks import ClassyHook

        assert isinstance(hooks, list)
        assert all(isinstance(hook, ClassyHook) for hook in hooks)
        assert len({hook.name() for hook in hooks}) == len(
            hooks
        ), "Cannot have repeated hooks of the same class"

        self.hooks = hooks
        return self

    def set_model(self, model: ClassyModel):
        """Set model for task

        Args:
            model: Model to be trained
        """
        self.base_model = model
        return self

    def set_test_only(self, test_only: bool):
        """Set test only flag

        Args:
            test_only: If true, only test phases will be run
        """
        self.test_only = test_only
        return self

    def set_amp_opt_level(self, opt_level: Optional[str]):
        """Disable / enable apex.amp and set the automatic mixed precision level.

        apex.amp can be utilized for mixed / half precision training.

        Args:
            opt_level: Opt level used to initialize apex.amp. Set to None to disable
                amp. Supported modes are -
                    O0: FP32 training
                    O1: Mixed Precision
                    O2: "Almost FP16" Mixed Precision
                    O3: FP16 training
                See https://nvidia.github.io/apex/amp.html#opt-levels for more info.
        Raises:
            RuntimeError: If opt_level is not None and apex is not installed.

        Warning: apex needs to be installed to utilize this feature.
        """
        if opt_level not in [None, "O0", "O1", "O2", "O3"]:
            raise ValueError(f"Unsupported opt_level: {opt_level}")

        self.amp_opt_level = opt_level

        if opt_level is None:
            logging.info(f"AMP disabled")
        else:
            if not apex_available:
                raise RuntimeError("apex is not installed, cannot enable amp")

            logging.info(f"AMP enabled with opt_level {opt_level}")
        return self

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ClassificationTask":
        """Instantiates a ClassificationTask from a configuration.

        Args:
            config: A configuration for a ClassificationTask.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A ClassificationTask instance.
        """
        optimizer_config = config["optimizer"]
        optimizer_config["num_epochs"] = config["num_epochs"]

        datasets = {}
        phase_types = ["train", "test"]
        for phase_type in phase_types:
            datasets[phase_type] = build_dataset(config["dataset"][phase_type])
        loss = build_loss(config["loss"])
        test_only = config.get("test_only", False)
        amp_opt_level = config.get("amp_opt_level")
        meters = build_meters(config.get("meters", {}))
        model = build_model(config["model"])
        # put model in eval mode in case any hooks modify model states, it'll
        # be reset to train mode before training
        model.eval()
        optimizer = build_optimizer(optimizer_config)

        task = (
            cls()
            .set_num_epochs(config["num_epochs"])
            .set_test_phase_period(config.get("test_phase_period", 1))
            .set_loss(loss)
            .set_test_only(test_only)
            .set_model(model)
            .set_optimizer(optimizer)
            .set_meters(meters)
            .set_amp_opt_level(amp_opt_level)
            .set_distributed_options(
                BroadcastBuffersMode[config.get("broadcast_buffers", "DISABLED")]
            )
        )
        for phase_type in phase_types:
            task.set_dataset(datasets[phase_type], phase_type)

        return task

    @property
    def num_batches_per_phase(self):
        """Returns number of batches in current phase iterator
        """
        return len(self.data_iterator)

    @property
    def model(self):
        """Returns model used in training (can be wrapped with DDP)
        """
        return (
            self.distributed_model if is_distributed_training_run() else self.base_model
        )

    @property
    def phase_type(self):
        """Returns current phase type. String with value "train" or "test"
        """
        return "train" if self.train else "test"

    @property
    def eval_phase_idx(self):
        """Returns current evaluation phase
        """
        return self.phase_idx - self.train_phase_idx - 1

    def get_data_iterator(self):
        """Returns data iterator for current phase
        """
        return self.data_iterator

    def get_total_training_phases(self):
        """
        Returns the total number of "train" phases in the task
        """
        num_training_phases = 0
        for phase in self.phases:
            if phase["train"] is True:
                num_training_phases += 1
        return num_training_phases

    def get_total_test_phases(self):
        """
        Returns the total number of "test" phases in the task
        """
        num_test_phases = 0
        for phase in self.phases:
            if phase["train"] is False:
                num_test_phases += 1
        return num_test_phases

    def _build_phases(self):
        """Returns list of phases from config.

        These phases will look like:
        {
          train: is this a train or test phase?
          optimizer: optimizer settings
        }

        If this is a test only run, then only test phases will be
        generated, if this is a training run, then x phases = x train
        phases + x test phases, interleaved.
        """
        if not self.test_only:
            phases = [{"train": True} for _ in range(self.num_epochs)]

            final_phases = []
            for i, phase in enumerate(phases):
                final_phases.append(phase)
                if (i + 1) % self.test_phase_period == 0:
                    final_phases.append({"train": False})
            if final_phases[-1]["train"]:
                final_phases.append({"train": False})
            return final_phases

        return [{"train": False} for _ in range(self.num_epochs)]

    def build_dataloader(
        self,
        phase_type,
        num_workers,
        pin_memory,
        multiprocessing_context=None,
        **kwargs,
    ):
        """Buildss a dataloader iterable for a particular phase type.

        Args:
            phase_type: "train" or "test" iterable
            num_workers: Number of dataloading processes. If 0,
                dataloading is done on main process. See `PyTorch dataloader
                documentation <https://pytorch.org/docs/stable/
                data.html#torch.utils.data.DataLoader>`_ for more details on
                ``num_workers`` and the usage
                of python multiprocessing in dataloaders
            pin_memory: if true pin memory on GPU. See PyTorch dataloader
                documentation for details on ``pin_memory``.
            multiprocessing_context: Determines how processes are spawned.
                Value must be one of None, "spawn", "fork", "forkserver".
                If None, then context is inherited from parent process

        Returns:
            Returns a iterable over the dataset
        """
        return self.datasets[phase_type].iterator(
            num_workers=num_workers,
            pin_memory=pin_memory,
            multiprocessing_context=multiprocessing_context,
            **kwargs,
        )

    def build_dataloaders(
        self, num_workers, pin_memory, multiprocessing_context=None, **kwargs
    ):
        """Build a dataloader for each phase type

        Args:
            num_workers: Number of dataloading processes. If 0,
                dataloading is done on main process. See `PyTorch dataloader
                documentation <https://pytorch.org/docs/stable/
                data.html#torch.utils.data.DataLoader>`_
                for more details on num_workers and the usage
                of python multiprocessing in dataloaders
            pin_memory: if true pin memory on GPU. See PyTorch dataloader
                documentation for details on pin_memory.
            multiprocessing_context: Determines how processes are spawned.
                Value must be one of None, "spawn", "fork", "forkserver".
                If None, then context is inherited from parent process

        Returns:
            Returns an iterable over the dataset associated with each phase_type
        """
        return {
            phase_type: self.build_dataloader(
                phase_type,
                num_workers=num_workers,
                pin_memory=pin_memory,
                multiprocessing_context=multiprocessing_context,
                **kwargs,
            )
            for phase_type in self.datasets.keys()
        }

    def prepare(
        self,
        num_dataloader_workers=0,
        pin_memory=False,
        use_gpu=False,
        dataloader_mp_context=None,
    ):
        """Prepares task for training, populates all derived attributes

        Args:
            num_dataloader_workers: Number of dataloading processes. If 0,
                dataloading is done on main process
            pin_memory: if true pin memory on GPU
            use_gpu: if true, load model, optimizer, loss, etc on GPU
            dataloader_mp_context: Determines how processes are spawned.
                Value must be one of None, "spawn", "fork", "forkserver".
                If None, then context is inherited from parent process
        """
        self.phases = self._build_phases()
        self.dataloaders = self.build_dataloaders(
            num_workers=num_dataloader_workers,
            pin_memory=pin_memory,
            multiprocessing_context=dataloader_mp_context,
        )

        # move the model and loss to the right device
        if use_gpu:
            self.base_model, self.loss = copy_model_to_gpu(self.base_model, self.loss)
        else:
            self.loss.cpu()
            self.base_model.cpu()

        # initialize the pytorch optimizer now since the model has been moved to
        # the appropriate device
        self.optimizer.init_pytorch_optimizer(self.base_model, loss=self.loss)

        classy_state_dict = (
            None
            if self.checkpoint is None
            else self.checkpoint.get("classy_state_dict")
        )

        if classy_state_dict is not None:
            state_load_success = update_classy_state(self, classy_state_dict)
            assert (
                state_load_success
            ), "Update classy state from checkpoint was unsuccessful."

        if self.amp_opt_level is not None:
            # Initialize apex.amp. This updates the model and the PyTorch optimizer (
            # which is wrapped by the ClassyOptimizer in self.optimizer)
            self.base_model, self.optimizer.optimizer = apex.amp.initialize(
                self.base_model, self.optimizer.optimizer, opt_level=self.amp_opt_level
            )
        self.init_distributed_data_parallel_model()

    def init_distributed_data_parallel_model(self):
        """
        Initialize
        `torch.nn.parallel.distributed.DistributedDataParallel <https://pytorch.org/
        docs/stable/nn.html#distributeddataparallel>`_.

        Needed for distributed training. This is where a model should be wrapped by DDP.
        """
        if not is_distributed_training_run():
            return
        assert (
            self.distributed_model is None
        ), "init_ddp_non_elastic must only be called once"

        broadcast_buffers = (
            self.broadcast_buffers_mode == BroadcastBuffersMode.FORWARD_PASS
        )
        self.distributed_model = init_distributed_data_parallel_model(
            self.base_model, broadcast_buffers=broadcast_buffers
        )
        if isinstance(self.loss, ClassyLoss) and self.loss.has_learned_parameters():
            logging.info("Initializing distributed loss")
            self.loss = init_distributed_data_parallel_model(
                self.loss, broadcast_buffers=broadcast_buffers
            )

    @property
    def where(self):
        """Returns the proportion of training that has completed. If in test
        only mode, returns proportion of testing completed

        Returned value is a float in the range [0, 1)
        """
        current_step = self.num_updates / self.get_global_batchsize()
        num_phases = (
            self.get_total_test_phases()
            if self.test_only
            else self.get_total_training_phases()
        )

        if self.num_batches_per_phase <= 0:
            raise RuntimeError("No batches to read. Is the dataset empty?")

        num_steps = num_phases * self.num_batches_per_phase
        where = current_step / num_steps

        assert where >= 0 and where < 1, f"Invalid where: {where}"

        return where

    def get_classy_state(self, deep_copy: bool = False):
        """Returns serialiable state of task

        Args:
            deep_copy: If true, does a deep copy of state before returning.
        """
        classy_state_dict = {
            "train": self.train,
            "base_model": self.base_model.get_classy_state(),
            "meters": [meter.get_classy_state() for meter in self.meters],
            "optimizer": self.optimizer.get_classy_state(),
            "phase_idx": self.phase_idx,
            "train_phase_idx": self.train_phase_idx,
            "num_updates": self.num_updates,
            "losses": self.losses,
            "hooks": {hook.name(): hook.get_classy_state() for hook in self.hooks},
            "loss": {},
        }
        if isinstance(self.loss, ClassyLoss):
            classy_state_dict["loss"] = self.loss.get_classy_state()
        if deep_copy:
            classy_state_dict = copy.deepcopy(classy_state_dict)
        return classy_state_dict

    def set_classy_state(self, state):
        """Set task state

        Args:
            state: Dict containing state of a task
        """
        # some settings are different in test only
        self.train = False if self.test_only else state["train"]
        if not self.test_only:
            self.phase_idx = state["phase_idx"]
            self.num_updates = state["num_updates"]
            self.train_phase_idx = state["train_phase_idx"]
            self.losses = state["losses"]
            for meter, meter_state in zip(self.meters, state["meters"]):
                meter.set_classy_state(meter_state)

        self.base_model.set_classy_state(state["base_model"])
        self.optimizer.set_classy_state(state["optimizer"])
        if state.get("loss") and isinstance(self.loss, ClassyLoss):
            self.loss.set_classy_state(state["loss"])

        for hook in self.hooks:
            # we still want to be able to run when new hooks are added or old
            # hooks are removed
            if hook.name() in state["hooks"]:
                hook.set_classy_state(state["hooks"][hook.name()])
            else:
                logging.warn(f"No state found for hook: {hook.name()}")
        # TODO (mannatsingh): Figure out how to set the state of the dataloaders
        # Re-build dataloader & re-create iterator.
        self._recreate_data_loader_from_dataset()
        self.create_data_iterator()
        # Set up pytorch module in train vs eval mode, update optimizer.
        self._set_model_train_mode()

    def eval_step(self, use_gpu, local_variables=None):
        if local_variables is None:
            local_variables = {}

        # Process next sample
        sample = next(self.get_data_iterator())
        local_variables["sample"] = sample

        assert (
            isinstance(local_variables["sample"], dict)
            and "input" in local_variables["sample"]
            and "target" in local_variables["sample"]
        ), (
            f"Returned sample [{sample}] is not a map with 'input' and"
            + "'target' keys"
        )

        # Copy sample to GPU
        local_variables["target"] = local_variables["sample"]["target"]
        if use_gpu:
            for key, value in local_variables["sample"].items():
                local_variables["sample"][key] = recursive_copy_to_gpu(
                    value, non_blocking=True
                )

        with torch.no_grad():
            local_variables["output"] = self.model(local_variables["sample"]["input"])

            local_variables["local_loss"] = self.compute_loss(
                local_variables["output"], local_variables["sample"]
            )

            local_variables["loss"] = local_variables["local_loss"].detach().clone()
            local_variables["loss"] = all_reduce_mean(local_variables["loss"])

            self.losses.append(
                local_variables["loss"].data.cpu().item()
                * local_variables["target"].size(0)
            )

            self.update_meters(local_variables["output"], local_variables["sample"])

    def train_step(self, use_gpu, local_variables=None):
        """Train step to be executed in train loop

        Args:
            use_gpu: if true, execute training on GPU
            local_variables: Dict containing intermediate values
                in train_step for access by hooks
        """

        if local_variables is None:
            local_variables = {}

        # Process next sample
        sample = next(self.get_data_iterator())
        local_variables["sample"] = sample

        assert (
            isinstance(local_variables["sample"], dict)
            and "input" in local_variables["sample"]
            and "target" in local_variables["sample"]
        ), (
            f"Returned sample [{sample}] is not a map with 'input' and"
            + "'target' keys"
        )

        # Copy sample to GPU
        local_variables["target"] = local_variables["sample"]["target"]
        if use_gpu:
            for key, value in local_variables["sample"].items():
                local_variables["sample"][key] = recursive_copy_to_gpu(
                    value, non_blocking=True
                )

        with torch.enable_grad():
            # Forward pass
            local_variables["output"] = self.model(local_variables["sample"]["input"])

            local_variables["local_loss"] = self.compute_loss(
                local_variables["output"], local_variables["sample"]
            )

            local_variables["loss"] = local_variables["local_loss"].detach().clone()
            local_variables["loss"] = all_reduce_mean(local_variables["loss"])

            self.losses.append(
                local_variables["loss"].data.cpu().item()
                * local_variables["target"].size(0)
            )

            self.update_meters(local_variables["output"], local_variables["sample"])

        # Run backwards pass / update optimizer
        if self.amp_opt_level is not None:
            self.optimizer.zero_grad()
            with apex.amp.scale_loss(
                local_variables["local_loss"], self.optimizer.optimizer
            ) as scaled_loss:
                scaled_loss.backward()
        else:
            self.optimizer.backward(local_variables["local_loss"])

        self.optimizer.update_schedule_on_step(self.where)
        self.optimizer.step()

        self.num_updates += self.get_global_batchsize()

    def compute_loss(self, model_output, sample):
        return self.loss(model_output, sample["target"])

    def update_meters(self, model_output, sample):
        target = sample["target"].detach().cpu()
        model_output = model_output.detach().cpu()

        # Update meters
        for meter in self.meters:
            meter.update(model_output, target, is_train=self.train)

    def advance_phase(self):
        """Performs bookkeeping / task updates between phases

        Increments phase idx, resets meters, resets loss history,
        resets counters, shuffles dataset, rebuilds iterators, and
        sets the train / test state for phase.
        """
        logging.info("Advancing phase")
        # Reset meters for next phase / epoch
        for meter in self.meters:
            meter.reset()

        # Reset loss history for next epoch
        self.losses = []

        # Setup new phase
        self.phase_idx += 1
        phase = self.phases[self.phase_idx]
        self.train = True if phase["train"] else False
        if self.train:
            self.train_phase_idx += 1

        # Re-build dataloader & re-create iterator anytime membership changes.
        self._recreate_data_loader_from_dataset()
        self.create_data_iterator()
        # Set up pytorch module in train vs eval mode, update optimizer.
        self._set_model_train_mode()

        # Update the optimizer schedule
        if self.train and self.train_phase_idx >= 0:
            self.optimizer.update_schedule_on_epoch(self.where)

    def done_training(self):
        """Stop condition for training
        """
        return self.phase_idx + 1 >= len(self.phases)

    def _recreate_data_loader_from_dataset(self, phase_type=None):
        """
        This utility is invoked to re-create the data loader object
        for the current phase of execution, using the existing dataset.
        This is sufficient when advancing phases.
        """
        if phase_type is None:
            phase_type = self.phase_type

        logging.info("Recreating data loader for new phase")
        num_workers = 0
        if hasattr(self.dataloaders[phase_type], "num_workers"):
            num_workers = self.dataloaders[phase_type].num_workers
        pin_memory = False
        if hasattr(self.dataloaders[phase_type], "pin_memory"):
            pin_memory = self.dataloaders[phase_type].pin_memory
        multiprocessing_context = None
        if hasattr(self.dataloaders[phase_type], "multiprocessing_context"):
            multiprocessing_context = self.dataloaders[
                phase_type
            ].multiprocessing_context
        if phase_type == "test":
            current_phase_id = 0
        else:
            current_phase_id = max(self.train_phase_idx, 0)

        self.dataloaders[phase_type] = self.build_dataloader(
            phase_type=phase_type,
            num_workers=num_workers,
            pin_memory=pin_memory,
            multiprocessing_context=multiprocessing_context,
            current_phase_id=current_phase_id,
        )

    def create_data_iterator(self):
        """Creates data iterator for phase.
        """
        # Delete iterator explicitly so that all dataloader processes
        # are cleaned up.
        del self.data_iterator
        self.data_iterator = iter(self.dataloaders[self.phase_type])

    def _set_model_train_mode(self):
        """Set train mode for model
        """
        phase = self.phases[self.phase_idx]
        self.base_model.train(phase["train"])
        self.loss.train(phase["train"])

        if (
            self.broadcast_buffers_mode == BroadcastBuffersMode.BEFORE_EVAL
            and not self.train
        ):
            self._broadcast_buffers()

    def _broadcast_buffers(self):
        """Explicitly synchronize buffers across all devices."""
        if self.distributed_model is None:
            return
        buffers = list(self.base_model.buffers())
        if len(buffers) > 0:
            logging.info("Synchronizing buffers before evaluation.")
            for buffer in buffers:
                broadcast(buffer, 0, group=self.distributed_model.process_group)

    # TODO: Functions below should be better abstracted into the dataloader
    # abstraction
    def get_batchsize_per_replica(self):
        """Return local replica's batchsize for dataset (e.g. batchsize per GPU)
        """
        # TODO(T47573564) - cleaner abstraction
        return self.dataloaders[self.phase_type].dataset.get_batchsize_per_replica()

    def get_global_batchsize(self):
        """Return global batchsize across all trainers
        """
        return self.dataloaders[self.phase_type].dataset.get_global_batchsize()

    def on_start(self, local_variables):
        self.run_hooks(local_variables, ClassyHookFunctions.on_start.name)

    def on_phase_start(self, local_variables):
        self.phase_start_time_total = time.perf_counter()

        self.advance_phase()

        self.run_hooks(local_variables, ClassyHookFunctions.on_phase_start.name)

        self.phase_start_time_train = time.perf_counter()

    def on_phase_end(self, local_variables):
        self.log_phase_end("train")

        logging.info("Syncing meters on phase end...")
        for meter in self.meters:
            meter.sync_state()
        logging.info("...meters synced")
        barrier()

        self.run_hooks(local_variables, ClassyHookFunctions.on_phase_end.name)
        self.perf_log = []

        self.log_phase_end("total")

    def on_end(self, local_variables):
        self.run_hooks(local_variables, ClassyHookFunctions.on_end.name)

    def log_phase_end(self, tag):
        if not self.train:
            return

        start_time = (
            self.phase_start_time_train
            if tag == "train"
            else self.phase_start_time_total
        )
        phase_duration = time.perf_counter() - start_time
        im_per_sec = (
            self.get_global_batchsize() * self.num_batches_per_phase
        ) / phase_duration
        self.perf_log.append(
            {
                "tag": tag,
                "phase_idx": self.train_phase_idx,
                "epoch_duration": phase_duration,
                "im_per_sec": im_per_sec,
            }
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Implementation of ResNeXt (https://arxiv.org/pdf/1611.05431.pdf)
"""

import math
from typing import Any, Dict, List, Optional, Tuple, Union

import torch.nn as nn
from classy_vision.generic.util import is_pos_int

from . import register_model
from .classy_model import ClassyModel


# global setting for in-place ReLU:
INPLACE = True


def conv3x3(in_planes, out_planes, stride=1, groups=1):
    """helper function for constructing 3x3 grouped convolution"""
    return nn.Conv2d(
        in_planes,
        out_planes,
        kernel_size=3,
        stride=stride,
        padding=1,
        groups=groups,
        bias=False,
    )


def conv1x1(in_planes, out_planes, stride=1):
    """helper function for constructing 1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)


class GenericLayer(nn.Module):
    """
        Parent class for 2-layer (BasicLayer) and 3-layer (BottleneckLayer)
        bottleneck layer class
    """

    def __init__(
        self,
        convolutional_block,
        in_planes,
        out_planes,
        stride=1,
        mid_planes_and_cardinality=None,
        reduction=4,
        final_bn_relu=True,
    ):

        # assertions on inputs:
        assert is_pos_int(in_planes) and is_pos_int(out_planes)
        assert is_pos_int(stride) and is_pos_int(reduction)

        # set object fields:
        super(GenericLayer, self).__init__()
        self.convolutional_block = convolutional_block
        self.final_bn_relu = final_bn_relu

        # final batchnorm and relu layer:
        if final_bn_relu:
            self.bn = nn.BatchNorm2d(out_planes)
            self.relu = nn.ReLU(inplace=INPLACE)

        # define down-sampling layer (if direct residual impossible):
        self.downsample = None
        if stride != 1 or in_planes != out_planes:
            self.downsample = nn.Sequential(
                conv1x1(in_planes, out_planes, stride=stride),
                nn.BatchNorm2d(out_planes),
            )

    def forward(self, x):

        # if required, perform downsampling along shortcut connection:
        if self.downsample is None:
            residual = x
        else:
            residual = self.downsample(x)

        # forward pass through convolutional block:
        out = self.convolutional_block(x)

        if self.final_bn_relu:
            out = self.bn(out)
        # add residual connection, perform rely + batchnorm, and return result:
        out += residual
        if self.final_bn_relu:
            out = self.relu(out)
        return out


class BasicLayer(GenericLayer):
    """
        ResNeXt bottleneck layer with `in_planes` input planes and `out_planes`
        output planes.
    """

    def __init__(
        self,
        in_planes,
        out_planes,
        stride=1,
        mid_planes_and_cardinality=None,
        reduction=4,
        final_bn_relu=True,
    ):

        # assertions on inputs:
        assert is_pos_int(in_planes) and is_pos_int(out_planes)
        assert is_pos_int(stride) and is_pos_int(reduction)

        # define convolutional block:
        convolutional_block = nn.Sequential(
            conv3x3(in_planes, out_planes, stride=stride),
            nn.BatchNorm2d(out_planes),
            nn.ReLU(inplace=INPLACE),
            conv3x3(out_planes, out_planes),
        )

        # call constructor of generic layer:
        super(BasicLayer, self).__init__(
            convolutional_block,
            in_planes,
            out_planes,
            stride=stride,
            reduction=reduction,
            final_bn_relu=final_bn_relu,
        )


class BottleneckLayer(GenericLayer):
    """
        ResNeXt bottleneck layer with `in_planes` input planes, `out_planes`
        output planes, and a bottleneck `reduction`.
    """

    def __init__(
        self,
        in_planes,
        out_planes,
        stride=1,
        mid_planes_and_cardinality=None,
        reduction=4,
        final_bn_relu=True,
    ):

        # assertions on inputs:
        assert is_pos_int(in_planes) and is_pos_int(out_planes)
        assert is_pos_int(stride) and is_pos_int(reduction)

        # define convolutional layers:
        bottleneck_planes = int(math.ceil(out_planes / reduction))
        cardinality = 1
        if mid_planes_and_cardinality is not None:
            mid_planes, cardinality = mid_planes_and_cardinality
            bottleneck_planes = mid_planes * cardinality

        convolutional_block = nn.Sequential(
            conv1x1(in_planes, bottleneck_planes),
            nn.BatchNorm2d(bottleneck_planes),
            nn.ReLU(inplace=INPLACE),
            conv3x3(
                bottleneck_planes, bottleneck_planes, stride=stride, groups=cardinality
            ),
            nn.BatchNorm2d(bottleneck_planes),
            nn.ReLU(inplace=INPLACE),
            conv1x1(bottleneck_planes, out_planes),
        )

        # call constructor of generic layer:
        super(BottleneckLayer, self).__init__(
            convolutional_block,
            in_planes,
            out_planes,
            stride=stride,
            reduction=reduction,
            final_bn_relu=final_bn_relu,
        )


class SmallInputInitialBlock(nn.Module):
    """
        ResNeXt initial block for small input with `in_planes` input planes
    """

    def __init__(self, init_planes):
        super().__init__()
        self._module = nn.Sequential(
            conv3x3(3, init_planes, stride=1),
            nn.BatchNorm2d(init_planes),
            nn.ReLU(inplace=INPLACE),
        )

    def forward(self, x):
        return self._module(x)


class InitialBlock(nn.Module):
    """
        ResNeXt initial block with `in_planes` input planes
    """

    def __init__(self, init_planes):
        super().__init__()
        self._module = nn.Sequential(
            nn.Conv2d(3, init_planes, kernel_size=7, stride=2, padding=3, bias=False),
            nn.BatchNorm2d(init_planes),
            nn.ReLU(inplace=INPLACE),
            nn.MaxPool2d(kernel_size=3, stride=2, padding=1),
        )

    def forward(self, x):
        return self._module(x)


@register_model("resnext")
class ResNeXt(ClassyModel):
    def __init__(
        self,
        num_blocks,
        init_planes: int = 64,
        reduction: int = 4,
        small_input: bool = False,
        zero_init_bn_residuals: bool = False,
        base_width_and_cardinality: Optional[Union[Tuple, List]] = None,
        basic_layer: bool = False,
        final_bn_relu: bool = True,
        bn_weight_decay: Optional[bool] = False,
    ):
        """
            Implementation of `ResNeXt <https://arxiv.org/pdf/1611.05431.pdf>`_.

            Set ``small_input`` to `True` for 32x32 sized image inputs.

            Set ``final_bn_relu`` to `False` to exclude the final batchnorm and
            ReLU layers. These settings are useful when training Siamese networks.
        """
        super().__init__()

        # assertions on inputs:
        assert type(num_blocks) == list
        assert all(is_pos_int(n) for n in num_blocks)
        assert is_pos_int(init_planes) and is_pos_int(reduction)
        assert type(small_input) == bool
        assert type(bn_weight_decay) == bool
        assert (
            type(zero_init_bn_residuals) == bool
        ), "zero_init_bn_residuals must be a boolean, set to true if gamma of last\
             BN of residual block should be initialized to 0.0, false for 1.0"
        assert base_width_and_cardinality is None or (
            isinstance(base_width_and_cardinality, (tuple, list))
            and len(base_width_and_cardinality) == 2
            and is_pos_int(base_width_and_cardinality[0])
            and is_pos_int(base_width_and_cardinality[1])
        )

        # Chooses whether to apply weight decay to batch norm
        # parameters. This improves results in some situations,
        # e.g. ResNeXt models trained / evaluated using the Imagenet
        # dataset, but can cause worse performance in other scenarios
        self.bn_weight_decay = bn_weight_decay

        # initial convolutional block:
        self.num_blocks = num_blocks
        self.small_input = small_input
        self._make_initial_block(small_input, init_planes, basic_layer)

        # compute number of planes at each spatial resolution:
        out_planes = [init_planes * 2 ** i * reduction for i in range(len(num_blocks))]
        in_planes = [init_planes] + out_planes[:-1]

        # create subnetworks for each spatial resolution:
        blocks = []
        for idx in range(len(out_planes)):
            mid_planes_and_cardinality = None
            if base_width_and_cardinality is not None:
                w, c = base_width_and_cardinality
                mid_planes_and_cardinality = (w * 2 ** idx, c)
            new_block = self._make_resolution_block(
                in_planes[idx],
                out_planes[idx],
                idx,
                num_blocks[idx],  # num layers
                stride=1 if idx == 0 else 2,
                mid_planes_and_cardinality=mid_planes_and_cardinality,
                reduction=reduction,
                final_bn_relu=final_bn_relu or (idx != (len(out_planes) - 1)),
            )
            blocks.append(nn.Sequential(*new_block))
        self.blocks = nn.Sequential(*blocks)

        self.out_planes = out_planes[-1]
        self._num_classes = out_planes

        # initialize weights:
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Init BatchNorm gamma to 0.0 for last BN layer, it gets 0.2-0.3% higher
        # final val top1 for larger batch sizes.
        if zero_init_bn_residuals:
            for m in self.modules():
                if isinstance(m, GenericLayer):
                    if hasattr(m, "bn"):
                        nn.init.constant_(m.bn.weight, 0)

    def _make_initial_block(self, small_input, init_planes, basic_layer):
        if small_input:
            self.initial_block = SmallInputInitialBlock(init_planes)
            self.layer_type = BasicLayer
        else:
            self.initial_block = InitialBlock(init_planes)
            self.layer_type = BasicLayer if basic_layer else BottleneckLayer

    # helper function that creates ResNet blocks at single spatial resolution:
    def _make_resolution_block(
        self,
        in_planes,
        out_planes,
        resolution_idx,
        num_blocks,
        stride=1,
        mid_planes_and_cardinality=None,
        reduction=4,
        final_bn_relu=True,
    ):

        # add the desired number of residual blocks:
        blocks = []
        for idx in range(num_blocks):
            blocks.append(
                self.build_attachable_block(
                    "block{}-{}".format(resolution_idx, idx),
                    self.layer_type(
                        in_planes if idx == 0 else out_planes,
                        out_planes,
                        stride=stride if idx == 0 else 1,  # only first block has stride
                        mid_planes_and_cardinality=mid_planes_and_cardinality,
                        reduction=reduction,
                        final_bn_relu=final_bn_relu or (idx != (num_blocks - 1)),
                    ),
                )
            )
        return blocks

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ResNeXt":
        """Instantiates a ResNeXt from a configuration.

        Args:
            config: A configuration for a ResNeXt.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A ResNeXt instance.
        """
        assert "num_blocks" in config
        config = {
            "num_blocks": config["num_blocks"],
            "init_planes": config.get("init_planes", 64),
            "reduction": config.get("reduction", 4),
            "base_width_and_cardinality": config.get("base_width_and_cardinality"),
            "small_input": config.get("small_input", False),
            "basic_layer": config.get("basic_layer", False),
            "final_bn_relu": config.get("final_bn_relu", True),
            "zero_init_bn_residuals": config.get("zero_init_bn_residuals", False),
            "bn_weight_decay": config.get("bn_weight_decay", False),
        }
        return cls(**config)

    # forward pass in residual network:
    def forward(self, x):
        # initial convolutional block:
        out = self.initial_block(x)

        # evaluate all residual blocks:
        # TODO: (kaizh) T43794289 exit early if there is no block that has heads
        self.blocks(out)

        # By default the classification layer is implemented as one head on top
        # of the last block. The head is automatically computed right after the
        # last block.
        head_outputs = self.execute_heads()
        if len(head_outputs) == 0:
            raise Exception("Expecting at least one head that generates output")
        elif len(head_outputs) == 1:
            return list(head_outputs.values())[0]
        else:
            return head_outputs

    def get_optimizer_params(self):
        return super().get_optimizer_params(bn_weight_decay=self.bn_weight_decay)

    @property
    def input_shape(self):
        if self.small_input:
            return (3, 32, 32)
        else:
            return (3, 224, 224)

    @property
    def output_shape(self):
        return (1, self._num_classes)

    @property
    def model_depth(self):
        return sum(self.num_blocks)


@register_model("resnet18")
class ResNet18(ResNeXt):
    def __init__(self):
        super().__init__(
            num_blocks=[2, 2, 2, 2], basic_layer=True, zero_init_bn_residuals=True
        )

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ResNeXt":
        return cls()


@register_model("resnet34")
class ResNet34(ResNeXt):
    def __init__(self):
        super().__init__(
            num_blocks=[3, 4, 6, 3], basic_layer=True, zero_init_bn_residuals=True
        )

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ResNeXt":
        return cls()


@register_model("resnet50")
class ResNet50(ResNeXt):
    def __init__(self):
        super().__init__(
            num_blocks=[3, 4, 6, 3], basic_layer=False, zero_init_bn_residuals=True
        )

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ResNeXt":
        return cls()


@register_model("resnet101")
class ResNet101(ResNeXt):
    def __init__(self):
        super().__init__(
            num_blocks=[3, 4, 23, 3], basic_layer=False, zero_init_bn_residuals=True
        )

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ResNeXt":
        return cls()


@register_model("resnet152")
class ResNet152(ResNeXt):
    def __init__(self):
        super().__init__(
            num_blocks=[3, 8, 36, 3], basic_layer=False, zero_init_bn_residuals=True
        )

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ResNeXt":
        return cls()


# Note, the ResNeXt models all have weight decay enabled for the batch
# norm parameters. We have found empirically that this gives better
# results when training on ImageNet (~0.5pp of top-1 acc) and brings
# our results on track with reported ImageNet results...but for
# training on other datasets, we have observed losses in accuracy (for
# example, the dataset used in https://arxiv.org/abs/1805.00932).
@register_model("resnext50_32x4d")
class ResNeXt50(ResNeXt):
    def __init__(self):
        super().__init__(
            num_blocks=[3, 4, 6, 3],
            basic_layer=False,
            zero_init_bn_residuals=True,
            base_width_and_cardinality=(4, 32),
            bn_weight_decay=True,
        )

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ResNeXt":
        return cls()


@register_model("resnext101_32x4d")
class ResNeXt101(ResNeXt):
    def __init__(self):
        super().__init__(
            num_blocks=[3, 4, 23, 3],
            basic_layer=False,
            zero_init_bn_residuals=True,
            base_width_and_cardinality=(4, 32),
            bn_weight_decay=True,
        )

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ResNeXt":
        return cls()


@register_model("resnext152_32x4d")
class ResNeXt152(ResNeXt):
    def __init__(self):
        super().__init__(
            num_blocks=[3, 8, 36, 3],
            basic_layer=False,
            zero_init_bn_residuals=True,
            base_width_and_cardinality=(4, 32),
            bn_weight_decay=True,
        )

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ResNeXt":
        return cls()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import unittest
from test.generic.optim_test_util import TestOptimizer

from classy_vision.optim.sgd import SGD


class TestSGDOptimizer(TestOptimizer, unittest.TestCase):
    def _get_config(self):
        return {
            "name": "sgd",
            "num_epochs": 90,
            "lr": 0.1,
            "momentum": 0.9,
            "weight_decay": 0.0001,
            "nesterov": False,
        }

    def _instance_to_test(self):
        return SGD



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch.nn as nn
import torchvision.models as models
from classy_vision.models import ClassyModel, register_model


@register_model("my_model")
class MyModel(ClassyModel):
    def __init__(self):
        super().__init__()
        self.model = nn.Sequential(
            nn.AdaptiveAvgPool2d((20, 20)),
            nn.Flatten(1),
            nn.Linear(3 * 20 * 20, 2),
            nn.Sigmoid(),
        )

    def forward(self, x):
        x = self.model(x)
        return x

    @classmethod
    def from_config(cls, config):
        return cls()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict

import torch.nn as nn


class ClassyLoss(nn.Module):
    """
    Base class to calculate the loss during training.

    This implementation of :class:`torch.nn.Module` allows building
    the loss object from a configuration file.
    """

    def __init__(self):
        """
        Constructor for ClassyLoss.
        """
        super(ClassyLoss, self).__init__()

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ClassyLoss":
        """Instantiates a ClassyLoss from a configuration.

        Args:
            config: A configuration for a ClassyLoss.

        Returns:
            A ClassyLoss instance.
        """
        raise NotImplementedError()

    def forward(self, output, target):
        """
        Compute the loss for the provided sample.

        Refer to :class:`torch.nn.Module` for more details.
        """
        raise NotImplementedError

    def get_optimizer_params(self, bn_weight_decay=False):
        """Gets optimizer params.

        The default implementation is very simple. Most losses have no learned
        parameters, so this is rarely needed.
        """
        params = [
            param for param in self.parameters(recurse=True) if param.requires_grad
        ]
        return {"regularized_params": params, "unregularized_params": []}

    def get_classy_state(self) -> Dict[str, Any]:
        """Get the state of the ClassyLoss.

        The returned state is used for checkpointing. Note that most losses are
        stateless and do not need to save any state.

        Returns:
            A state dictionary containing the state of the loss.
        """
        return self.state_dict()

    def set_classy_state(self, state: Dict[str, Any]) -> None:
        """Set the state of the ClassyLoss.

        Args:
            state_dict: The state dictionary. Must be the output of a call to
                :func:`get_classy_state`.

        This is used to load the state of the loss from a checkpoint. Note
        that most losses are stateless and do not need to load any state.
        """
        return self.load_state_dict(state)

    def has_learned_parameters(self) -> bool:
        """Does this loss have learned parameters?"""
        return any(
            len(params) > 0 for (_, params) in self.get_optimizer_params().items()
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
from typing import Any, Callable, Dict, Optional, Union

from classy_vision.dataset import ClassyDataset, register_dataset
from classy_vision.dataset.transforms import ClassyTransform, build_transforms
from torchvision.datasets import ImageFolder


@register_dataset("classy_imagenet")
class ImageNetDataset(ClassyDataset):
    def __init__(
        self,
        split: str,
        batchsize_per_replica: int,
        shuffle: bool,
        transform: Optional[Union[ClassyTransform, Callable]],
        num_samples: Optional[int],
        root: str,
    ):
        img_dir = os.path.join(root, split)
        dataset = ImageFolder(img_dir)
        super().__init__(
            dataset, batchsize_per_replica, shuffle, transform, num_samples
        )

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ImageNetDataset":
        """Instantiates a ImageNetDataset from a configuration.

        Args:
            config: A configuration for a ImageNetDataset.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A ImageNetDataset instance.
        """
        (
            transform_config,
            batchsize_per_replica,
            shuffle,
            num_samples,
        ) = cls.parse_config(config)
        split = config.get("split")
        root = config.get("root")

        transform = build_transforms(transform_config)
        return cls(
            split=split,
            batchsize_per_replica=batchsize_per_replica,
            shuffle=shuffle,
            transform=transform,
            num_samples=num_samples,
            root=root,
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import math
from typing import Any, Dict

from . import ClassyParamScheduler, register_param_scheduler


@register_param_scheduler("cosine")
class CosineParamScheduler(ClassyParamScheduler):
    """
    Changes the param value after every epoch based on a `cosine schedule <https:
    //arxiv.org/pdf/1608.03983.pdf>`_.
    Can be used for either cosine decay or cosine warmup schedules based on
    start and end values.

    Example:

        .. code-block:: python

          start_lr: 0.1
          end_lr: 0.0001
    """

    def __init__(self, start_lr: float, end_lr: float):
        super().__init__()
        self._start_lr = start_lr
        self._end_lr = end_lr

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "CosineParamScheduler":
        """Instantiates a CosineParamScheduler from a configuration.

        Args:
            config: A configuration for a CosineParamScheduler.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A CosineParamScheduler instance.
        """
        assert (
            "start_lr" in config and "end_lr" in config
        ), "Cosine scheduler requires a start_lr and a end_lr"

        return cls(start_lr=config["start_lr"], end_lr=config["end_lr"])

    def __call__(self, where: float):
        return self._end_lr + 0.5 * (self._start_lr - self._end_lr) * (
            1 + math.cos(math.pi * where)
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from hydra.plugins import SearchPathPlugin


class ClassyVisionPathPlugin(SearchPathPlugin):
    def manipulate_search_path(self, search_path):
        search_path.append("classy_vision", "pkg://classy_vision.hydra.conf")



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import copy
import unittest

from classy_vision.optim.param_scheduler import build_param_scheduler
from classy_vision.optim.param_scheduler.polynomial_decay_scheduler import (
    PolynomialDecayParamScheduler,
)


class TestPolynomialScheduler(unittest.TestCase):
    _num_epochs = 10

    def _get_valid_config(self):
        return {
            "name": "polynomial",
            "num_epochs": self._num_epochs,
            "base_lr": 0.1,
            "power": 1,
        }

    def test_invalid_config(self):
        # Invalid num epochs
        config = self._get_valid_config()

        # Invalid Base lr
        bad_config = copy.deepcopy(config)
        del bad_config["base_lr"]
        with self.assertRaises(AssertionError):
            PolynomialDecayParamScheduler.from_config(bad_config)

        # Invalid Power
        bad_config = copy.deepcopy(config)
        del bad_config["power"]
        with self.assertRaises(AssertionError):
            PolynomialDecayParamScheduler.from_config(bad_config)

    def test_scheduler(self):
        config = self._get_valid_config()

        scheduler = PolynomialDecayParamScheduler.from_config(config)
        schedule = [
            round(scheduler(epoch_num / self._num_epochs), 2)
            for epoch_num in range(self._num_epochs)
        ]
        expected_schedule = [0.1, 0.09, 0.08, 0.07, 0.06, 0.05, 0.04, 0.03, 0.02, 0.01]

        self.assertEqual(schedule, expected_schedule)

    def test_build_polynomial_scheduler(self):
        config = self._get_valid_config()
        scheduler = build_param_scheduler(config)
        self.assertTrue(isinstance(scheduler, PolynomialDecayParamScheduler))



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import bisect
import math
from typing import Any, Dict, List, NamedTuple, Optional, Union

from classy_vision.generic.util import is_pos_int

from . import ClassyParamScheduler, UpdateInterval, register_param_scheduler


@register_param_scheduler("multistep")
class MultiStepParamScheduler(ClassyParamScheduler):
    """
    Takes a predefined schedule for a param value, and a list of epochs
    which stand for the upper boundary (excluded) of each range.

    Example:

        .. code-block:: python

          values: [0.1, 0.01, 0.001, 0.0001]
          milestones = [30, 60, 80]
          num_epochs = 120

    Then the param value will be 0.1 for epochs 0-29, 0.01 for
    epochs 30-59, 0.001 for epochs 60-79, 0.0001 for epochs after epoch 80.
    Note that the length of values must be equal to the length of milestones
    plus one.
    """

    def __init__(
        self,
        values,
        num_epochs: int,
        update_interval: UpdateInterval,
        milestones: Optional[List[int]] = None,
    ):
        super().__init__(update_interval)
        self._param_schedule = values
        self._num_epochs = num_epochs
        self._milestones = milestones

        if milestones is None:
            # Default equispaced drop_epochs behavior
            self._milestones = []
            step_width = math.ceil(self._num_epochs / float(len(self._param_schedule)))
            for idx in range(len(self._param_schedule) - 1):
                self._milestones.append(step_width * (idx + 1))

        start_epoch = 0
        for milestone in self._milestones:
            # Do not exceed the total number of epochs
            assert milestone < self._num_epochs, (
                "Epoch milestone must be smaller than total number of epochs: num_epochs=%d, milestone=%d"
                % (self._num_epochs, milestone)
            )
            # Must be in ascending order
            assert start_epoch < milestone, (
                "Epoch milestone must be smaller than start epoch: start_epoch=%d, milestone=%d"
                % (start_epoch, milestone)
            )
            start_epoch = milestone

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "MultiStepParamScheduler":
        """Instantiates a MultiStepParamScheduler from a configuration.

        Args:
            config: A configuration for a MultiStepParamScheduler.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A MultiStepParamScheduler instance.
        """
        assert (
            "values" in config
            and isinstance(config["values"], list)
            and len(config["values"]) > 0
        ), "Non-Equi Step scheduler requires a list of at least one param value"
        assert is_pos_int(config["num_epochs"]), "Num epochs must be a positive integer"
        assert config["num_epochs"] >= len(
            config["values"]
        ), "Num epochs must be greater than param schedule"

        milestones = config.get("milestones", None)
        if "milestones" in config:
            assert (
                isinstance(config["milestones"], list)
                and len(config["milestones"]) == len(config["values"]) - 1
            ), (
                "Non-Equi Step scheduler requires a list of %d epochs"
                % (len(config["values"]) - 1)
            )
        return cls(
            values=config["values"],
            num_epochs=config["num_epochs"],
            update_interval=UpdateInterval(config.get("update_interval", "epoch")),
            milestones=milestones,
        )

    def __call__(self, where: float):
        epoch_num = int((where + self.WHERE_EPSILON) * self._num_epochs)
        return self._param_schedule[bisect.bisect_right(self._milestones, epoch_num)]



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict

import torch
import torch.nn as nn
from classy_vision.generic.util import is_pos_int, is_pos_int_list

from . import register_model
from .classy_model import ClassyModel, ClassyModelEvaluationMode
from .resnext3d_stage import ResStage
from .resnext3d_stem import R2Plus1DStem, ResNeXt3DStem


model_stems = {
    "r2plus1d_stem": R2Plus1DStem,
    "resnext3d_stem": ResNeXt3DStem,
    # For more types of model stem, add them below
}


class ResNeXt3DBase(ClassyModel):
    def __init__(
        self,
        input_key,
        input_planes,
        clip_crop_size,
        frames_per_clip,
        num_blocks,
        stem_name,
        stem_planes,
        stem_temporal_kernel,
        stem_spatial_kernel,
        stem_maxpool,
    ):
        """
        ResNeXt3DBase implements everything in ResNeXt3D model except the
        construction of 4 stages. See more details in ResNeXt3D.
        """
        super(ResNeXt3DBase, self).__init__()

        self._input_key = input_key
        self.input_planes = input_planes
        self.clip_crop_size = clip_crop_size
        self.frames_per_clip = frames_per_clip
        self.num_blocks = num_blocks

        assert stem_name in model_stems, "unknown stem: %s" % stem_name
        self.stem = model_stems[stem_name](
            stem_temporal_kernel,
            stem_spatial_kernel,
            input_planes,
            stem_planes,
            stem_maxpool,
        )

    @classmethod
    def _parse_config(cls, config):
        ret_config = {}
        required_args = [
            "input_planes",
            "clip_crop_size",
            "skip_transformation_type",
            "residual_transformation_type",
            "frames_per_clip",
            "num_blocks",
        ]
        for arg in required_args:
            assert arg in config, "resnext3d model requires argument %s" % arg
            ret_config[arg] = config[arg]

        # Default setting for model stem, which is considered as stage 0. Stage
        # index starts from 0 as implemented in ResStageBase._block_name() method.
        #   stem_planes: No. of output channles of conv op in stem
        #   stem_temporal_kernel: temporal size of conv op in stem
        #   stem_spatial_kernel: spatial size of conv op in stem
        #   stem_maxpool: by default, spatial maxpool op is disabled in stem
        ret_config.update(
            {
                "input_key": config.get("input_key", None),
                "stem_name": config.get("stem_name", "resnext3d_stem"),
                "stem_planes": config.get("stem_planes", 64),
                "stem_temporal_kernel": config.get("stem_temporal_kernel", 3),
                "stem_spatial_kernel": config.get("stem_spatial_kernel", 7),
                "stem_maxpool": config.get("stem_maxpool", False),
            }
        )
        # Default setting for model stages 1, 2, 3 and 4
        #   stage_planes: No. of output channel of 1st conv op in stage 1
        #   stage_temporal_kernel_basis: Basis of temporal kernel sizes for each of
        #       the stage.
        #   temporal_conv_1x1: if True, do temporal convolution in the fist
        #      1x1 Conv3d. Otherwise, do it in the second 3x3 Conv3d (default settting)
        #   stage_temporal_stride: temporal stride for each stage
        #   stage_spatial_stride: spatial stride for each stage
        #   num_groups: No. of groups in 2nd (group) conv in the residual transformation
        #   width_per_group: No. of channels per group in 2nd (group) conv in the
        #     residual transformation
        ret_config.update(
            {
                "stage_planes": config.get("stage_planes", 256),
                "stage_temporal_kernel_basis": config.get(
                    "stage_temporal_kernel_basis", [[3], [3], [3], [3]]
                ),
                "temporal_conv_1x1": config.get(
                    "temporal_conv_1x1", [False, False, False, False]
                ),
                "stage_temporal_stride": config.get(
                    "stage_temporal_stride", [1, 2, 2, 2]
                ),
                "stage_spatial_stride": config.get(
                    "stage_spatial_stride", [1, 2, 2, 2]
                ),
                "num_groups": config.get("num_groups", 1),
                "width_per_group": config.get("width_per_group", 64),
            }
        )
        # Default setting for model parameter initialization
        ret_config.update(
            {
                "zero_init_residual_transform": config.get(
                    "zero_init_residual_transform", False
                )
            }
        )
        assert is_pos_int_list(ret_config["num_blocks"])
        assert is_pos_int(ret_config["stem_planes"])
        assert is_pos_int(ret_config["stem_temporal_kernel"])
        assert is_pos_int(ret_config["stem_spatial_kernel"])
        assert type(ret_config["stem_maxpool"]) == bool
        assert is_pos_int(ret_config["stage_planes"])
        assert type(ret_config["stage_temporal_kernel_basis"]) == list
        assert all(
            is_pos_int_list(l) for l in ret_config["stage_temporal_kernel_basis"]
        )
        assert type(ret_config["temporal_conv_1x1"]) == list
        assert is_pos_int_list(ret_config["stage_temporal_stride"])
        assert is_pos_int_list(ret_config["stage_spatial_stride"])
        assert is_pos_int(ret_config["num_groups"])
        assert is_pos_int(ret_config["width_per_group"])
        return ret_config

    def _init_parameter(self, zero_init_residual_transform):
        for m in self.modules():
            if isinstance(m, nn.Conv3d):
                if (
                    hasattr(m, "final_transform_op")
                    and m.final_transform_op
                    and zero_init_residual_transform
                ):
                    nn.init.constant_(m.weight, 0)
                else:
                    nn.init.kaiming_normal_(
                        m.weight, mode="fan_out", nonlinearity="relu"
                    )
                if m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.BatchNorm3d) and m.affine:
                if (
                    hasattr(m, "final_transform_op")
                    and m.final_transform_op
                    and zero_init_residual_transform
                ):
                    batchnorm_weight = 0.0
                else:
                    batchnorm_weight = 1.0
                nn.init.constant_(m.weight, batchnorm_weight)
                nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, mean=0.0, std=0.01)
                nn.init.constant_(m.bias, 0)

    def set_classy_state(self, state):
        # We need to support both regular checkpoint loading and 2D conv weight
        # inflation into 3D conv weight in this function.
        self.load_head_states(state)
        current_state = self.state_dict()
        for name, weight_src in state["model"]["trunk"].items():
            assert name in current_state, (
                "weight %s is not found in ResNeXt3D model" % name
            )
            weight_tgt = current_state[name]
            assert (
                weight_src.dim() == weight_tgt.dim()
            ), "weight of source- and target 3D convolution should have same dimension"
            if (
                weight_src.dim() == 5
                and weight_src.shape[2] == 1
                and weight_tgt.shape[2] > 1
            ):
                # Find a source weight tensor where temporal dimension is 1. If the
                # temporal dimension of the current weight tensor with the same name
                # is larger than 1, we inflate the source weight tensor before
                # loading it. Such parameter inflation was first introduced in
                # the paper (https://arxiv.org/abs/1705.07750). It can achieve a
                # better initialization compared to random initialization.
                assert (
                    weight_src.shape[-2:] == weight_tgt.shape[-2:]
                    and weight_src.shape[:2] == weight_tgt.shape[:2]
                ), "weight shapes of source- and target 3D convolution mismatch"
                weight_src_inflated = (
                    weight_src.repeat(1, 1, weight_tgt.shape[2], 1, 1)
                    / weight_tgt.shape[2]
                )
                weight_src = weight_src_inflated
            else:
                assert all(
                    weight_src.size(d) == weight_tgt.size(d)
                    for d in range(weight_src.dim())
                ), (
                    "the shapes of source and target weight mismatch: %s Vs %s"
                    % (str(weight_src.size()), str(weight_tgt.size()))
                )

            current_state[name] = weight_src.clone()
        super().load_state_dict(current_state)

    def forward(self, x):
        """
        Args:
            x (dict or torch.Tensor): video input.
                When its type is dict, the dataset is a video dataset, and its
                content is like {"video": torch.tensor, "audio": torch.tensor}.
                When its type is torch.Tensor, the dataset is an image dataset.
        """
        assert isinstance(x, dict) or isinstance(
            x, torch.Tensor
        ), "x must be either a dictionary or a torch.Tensor"
        if isinstance(x, dict):
            assert self._input_key is not None and self._input_key in x, (
                "input key (%s) not in the input" % self._input_key
            )
            x = x[self._input_key]
        else:
            assert (
                self._input_key is None
            ), "when input of forward pass is a tensor, input key should not be set"
            assert x.dim() == 4 or x.dim() == 5, "tensor x must be 4D/5D tensor"
            if x.dim() == 4:
                # x is a 4D tensor of size N x C x H x W and is prepared from an
                # image dataset. We insert a temporal axis make it 5D of size
                # N x C x T x H x W
                x = torch.unsqueeze(x, 2)

        out = self.stem([x])
        out = self.stages(out)

        head_outputs = self.execute_heads()
        if len(head_outputs) == 0:
            raise Exception("Expecting at least one head that generates output")
        elif len(head_outputs) == 1:
            return list(head_outputs.values())[0]
        else:
            return head_outputs

    @property
    def input_shape(self):
        """
        Shape of video model input can vary in the following cases
        - At training stage, input are video frame croppings of fixed size.
        - At test stage, input are original video frames to support Fully Convolutional
            evaluation and its size can vary video by video
        """
        # Input shape is used by tensorboard hook. We put the input shape at
        # training stage for profiling and visualization purpose.
        return (
            self.input_planes,
            self.frames_per_clip,
            self.clip_crop_size,
            self.clip_crop_size,
        )

    @property
    def output_shape(self):
        return (1, None)

    @property
    def model_depth(self):
        return sum(self.num_blocks)

    @property
    def evaluation_mode(self):
        return ClassyModelEvaluationMode.VIDEO_CLIP_AVERAGING

    @property
    def input_key(self):
        return self._input_key


@register_model("resnext3d")
class ResNeXt3D(ResNeXt3DBase):
    """
    Implementation of:
        1. Conventional `post-activated 3D ResNe(X)t <https://arxiv.org/
        abs/1812.03982>`_.

        2. `Pre-activated 3D ResNe(X)t <https://arxiv.org/abs/1811.12814>`_.
        The model consists of one stem, a number of stages, and one or multiple
        heads that are attached to different blocks in the stage.
    """

    def __init__(
        self,
        input_key,
        input_planes,
        clip_crop_size,
        skip_transformation_type,
        residual_transformation_type,
        frames_per_clip,
        num_blocks,
        stem_name,
        stem_planes,
        stem_temporal_kernel,
        stem_spatial_kernel,
        stem_maxpool,
        stage_planes,
        stage_temporal_kernel_basis,
        temporal_conv_1x1,
        stage_temporal_stride,
        stage_spatial_stride,
        num_groups,
        width_per_group,
        zero_init_residual_transform,
    ):
        """
        Args:
            input_key (str): a key that can index into model input that is
                of dict type.
            input_planes (int): the channel dimension of the input. Normally 3 is used
                for rgb input.
            clip_crop_size (int): spatial cropping size of video clip at train time.
            skip_transformation_type (str): the type of skip transformation.
            residual_transformation_type (str): the type of residual transformation.
            frames_per_clip (int): Number of frames in a video clip.
            num_blocks (list): list of the number of blocks in stages.
            stem_name (str): name of model stem.
            stem_planes (int): the output dimension of the convolution in the model
                stem.
            stem_temporal_kernel (int): the temporal kernel size of the convolution
                in the model stem.
            stem_spatial_kernel (int): the spatial kernel size of the convolution
                in the model stem.
            stem_maxpool (bool): If true, perform max pooling.
            stage_planes (int): the output channel dimension of the 1st residual stage
            stage_temporal_kernel_basis (list): Basis of temporal kernel sizes for
                each of the stage.
            temporal_conv_1x1 (bool): Only useful for BottleneckTransformation.
                In a pathaway, if True, do temporal convolution in the first 1x1
                Conv3d. Otherwise, do it in the second 3x3 Conv3d.
            stage_temporal_stride (int): the temporal stride of the residual
                transformation.
            stage_spatial_stride (int): the spatial stride of the the residual
                transformation.
            num_groups (int): number of groups for the convolution.
                num_groups = 1 is for standard ResNet like networks, and
                num_groups > 1 is for ResNeXt like networks.
            width_per_group (int): Number of channels per group in 2nd (group)
                conv in the residual transformation in the first stage
            zero_init_residual_transform (bool): if true, the weight of last
                operation, which could be either BatchNorm3D in post-activated
                transformation or Conv3D in pre-activated transformation, in the
                residual transformation is initialized to zero
        """
        super(ResNeXt3D, self).__init__(
            input_key,
            input_planes,
            clip_crop_size,
            frames_per_clip,
            num_blocks,
            stem_name,
            stem_planes,
            stem_temporal_kernel,
            stem_spatial_kernel,
            stem_maxpool,
        )

        num_stages = len(num_blocks)
        out_planes = [stage_planes * 2 ** i for i in range(num_stages)]
        in_planes = [stem_planes] + out_planes[:-1]
        inner_planes = [
            num_groups * width_per_group * 2 ** i for i in range(num_stages)
        ]

        stages = []
        for s in range(num_stages):
            stage = ResStage(
                s + 1,  # stem is viewed as stage 0, and following stages start from 1
                [in_planes[s]],
                [out_planes[s]],
                [inner_planes[s]],
                [stage_temporal_kernel_basis[s]],
                [temporal_conv_1x1[s]],
                [stage_temporal_stride[s]],
                [stage_spatial_stride[s]],
                [num_blocks[s]],
                [num_groups],
                skip_transformation_type,
                residual_transformation_type,
                block_callback=self.build_attachable_block,
                disable_pre_activation=(s == 0),
                final_stage=(s == (num_stages - 1)),
            )
            stages.append(stage)

        self.stages = nn.Sequential(*stages)
        self._init_parameter(zero_init_residual_transform)

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "ResNeXt3D":
        """Instantiates a ResNeXt3D from a configuration.

        Args:
            config: A configuration for a ResNeXt3D.
                See :func:`__init__` for parameters expected in the config.

        Returns:
            A ResNeXt3D instance.
        """
        ret_config = ResNeXt3D._parse_config(config)
        return cls(**ret_config)
